import unittest

from kiss_headers import (
    Accept,
    AcceptEncoding,
    AcceptLanguage,
    AltSvc,
    CacheControl,
    Connection,
    ContentEncoding,
    ContentLength,
    ContentType,
    Date,
    Expires,
    Header,
    Host,
    IfModifiedSince,
    IfNoneMatch,
    Referer,
    Server,
    SetCookie,
    StrictTransportSecurity,
    UpgradeInsecureRequests,
    UserAgent,
    XFrameOptions,
    XXssProtection,
    parse_it,
)

RAW_HEADERS = """accept-ch: DPR
accept-ch-lifetime: 2592000
alt-svc: quic=":443"; ma=2592000; v="46,43", h3-Q050=":443"; ma=2592000, h3-Q049=":443"; ma=2592000, h3-Q048=":443"; ma=2592000, h3-Q046=":443"; ma=2592000, h3-Q043=":443"; ma=2592000
cache-control: private, max-age=0
content-encoding: br
content-length: 64032
content-type: text/html; charset=UTF-8
date: Mon, 16 Mar 2020 21:27:31 GMT
expires: -1
p3p: CP="This is not a P3P policy! See g.co/p3phelp for more info."
server: gws
set-cookie: 1P_JAR=2020-03-16-21; expires=Wed, 15-Apr-2020 21:27:31 GMT; path=/; domain=.google.fr; Secure; SameSite=none
set-cookie: NID=200=IGpBMMA3G7tki0niFFATFQ2BnsNceVP6XBtwOutoyw97AJ4_YFT5l1oLfLeX22xeI_STiP4omAB4rmMP3Sxgyo287ldQGwdZSdPOOZ_Md3roDOMAOtXEQ_hFbUvo0VPjS2gL1y00_6kQwpVxCghI2Ozrx-A4Xks3ZIXRj11RsWs; expires=Tue, 15-Sep-2020 21:27:31 GMT; path=/; domain=.google.fr; Secure; HttpOnly; SameSite=none
set-cookie: CONSENT=WP.284b10; expires=Fri, 01-Jan-2038 00:00:00 GMT; path=/; domain=.google.fr
status: 200
strict-transport-security: max-age=31536000
x-frame-options: SAMEORIGIN
x-xss-protection: 0"""

RAW_HEADERS_MOZILLA = """GET /home.html HTTP/1.1
Host: developer.mozilla.org
User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0
Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
Accept-Language: en-US,en;q=0.5
Accept-Encoding: gzip, deflate, br
Referer: https://developer.mozilla.org/testpage.html
Connection: keep-alive
Upgrade-Insecure-Requests: 1
If-Modified-Since: Mon, 18 Jul 2016 02:36:04 GMT
If-None-Match: "c561c68d0ba92bbeb8b0fff2a9199f722e3a621a"
Cache-Control: max-age=0"""


class MyBuilderCreationTestCase(unittest.TestCase):
    def test_replicate_raw_from_objects_request(self):
        headers = (
            Host("developer.mozilla.org")
            + UserAgent(
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0"
            )
            + Accept("text/html")
            + Accept("application/xhtml+xml")
            + Accept("application/xml", qualifier=0.9)
            + Accept(qualifier=0.8)
            + AcceptLanguage("en-US")
            + AcceptLanguage("en", qualifier=0.5)
            + AcceptEncoding("gzip")
            + AcceptEncoding("deflate")
            + AcceptEncoding("br")
            + Referer("https://developer.mozilla.org/testpage.html")
            + Connection(True)
            + UpgradeInsecureRequests()
            + IfModifiedSince("Mon, 18 Jul 2016 02:36:04 GMT")
            + IfNoneMatch("c561c68d0ba92bbeb8b0fff2a9199f722e3a621a")
            + CacheControl(max_age=0)
        )

        self.assertEqual(parse_it(RAW_HEADERS_MOZILLA), headers)

    def test_replicate_raw_from_objects(self):
        headers = (
            Header("Accept-Ch", "DPR")
            + Header("Accept-Ch-Lifetime", "2592000")
            + AltSvc("quic", ":443", max_age=2592000, versions=["46", "43"])
            + AltSvc("h3-Q050", ":443", max_age=2592000)
            + AltSvc("h3-Q049", ":443", max_age=2592000)
            + AltSvc("h3-Q048", ":443", max_age=2592000)
            + AltSvc("h3-Q046", ":443", max_age=2592000)
            + AltSvc("h3-Q043", ":443", max_age=2592000)
            + CacheControl("private")
            + CacheControl(max_age=0)
            + ContentEncoding("br")
            + ContentLength(64032)
            + ContentType("text/html", charset="utf-8")
            + Date("Mon, 16 Mar 2020 21:27:31 GMT")
            + Expires("-1")
            + Header(
                "P3P", 'CP="This is not a P3P policy! See g.co/p3phelp for more info."'
            )
            + Server("gws")
            + SetCookie(
                "1P_JAR",
                "2020-03-16-21",
                expires="Wed, 15-Apr-2020 21:27:31 GMT",
                path="/",
                domain=".google.fr",
                samesite="none",
                is_secure=True,
                is_httponly=False,
            )
            + SetCookie(
                "NID",
                "200=IGpBMMA3G7tki0niFFATFQ2BnsNceVP6XBtwOutoyw97AJ4_YFT5l1oLfLeX22xeI_STiP4omAB4rmMP3Sxgyo287ldQGwdZSdPOOZ_Md3roDOMAOtXEQ_hFbUvo0VPjS2gL1y00_6kQwpVxCghI2Ozrx-A4Xks3ZIXRj11RsWs",
                expires="Tue, 15-Sep-2020 21:27:31 GMT",
                path="/",
                domain=".google.fr",
                samesite="none",
                is_secure=True,
            )
            + SetCookie(
                "CONSENT",
                "WP.284b10",
                expires="Fri, 01-Jan-2038 00:00:00 GMT",
                path="/",
                domain=".google.fr",
                is_httponly=False,
            )
            + Header("Status", "200")
            + StrictTransportSecurity(max_age=31536000)
            + XFrameOptions("SAMEORIGIN")
            + XXssProtection(False)
        )

        self.assertEqual(parse_it(RAW_HEADERS), headers)


if __name__ == "__main__":
    unittest.main()



import unittest

from kiss_headers import Header, Headers, lock_output_type, parse_it
from kiss_headers.utils import decode_partials

RAW_HEADERS = """accept-ch: DPR
accept-ch-lifetime: 2592000
alt-svc: quic=":443"; ma=2592000; v="46,43", h3-Q050=":443"; ma=2592000, h3-Q049=":443"; ma=2592000, h3-Q048=":443"; ma=2592000, h3-Q046=":443"; ma=2592000, h3-Q043=":443"; ma=2592000
cache-control: private, max-age=0
content-encoding: br
content-length: 64032
content-type: text/html; charset=UTF-8
date: Mon, 16 Mar 2020 21:27:31 GMT
expires: -1
p3p: CP="This is not a P3P policy! See g.co/p3phelp for more info."
server: gws
set-cookie: 1P_JAR=2020-03-16-21; expires=Wed, 15-Apr-2020 21:27:31 GMT; path=/; domain=.google.fr; Secure; SameSite=none
set-cookie: NID=200=IGpBMMA3G7tki0niFFATFQ2BnsNceVP6XBtwOutoyw97AJ4_YFT5l1oLfLeX22xeI_STiP4omAB4rmMP3Sxgyo287ldQGwdZSdPOOZ_Md3roDOMAOtXEQ_hFbUvo0VPjS2gL1y00_6kQwpVxCghI2Ozrx-A4Xks3ZIXRj11RsWs; expires=Tue, 15-Sep-2020 21:27:31 GMT; path=/; domain=.google.fr; Secure; HttpOnly; SameSite=none
set-cookie: CONSENT=WP.284b10; expires=Fri, 01-Jan-2038 00:00:00 GMT; path=/; domain=.google.fr
status: 200
strict-transport-security: max-age=31536000
x-frame-options: SAMEORIGIN
x-xss-protection: 0""".replace(
    "\n", "\r\n"
)

RAW_HEADERS_MOZILLA = """GET /home.html HTTP/1.1
Host: developer.mozilla.org
User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0
Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
Accept-Language: en-US,en;q=0.5
Accept-Encoding: gzip, deflate, br
Referer: https://developer.mozilla.org/testpage.html
Connection: keep-alive
Upgrade-Insecure-Requests: 1
If-Modified-Since: Mon, 18 Jul 2016 02:36:04 GMT
If-None-Match: "c561c68d0ba92bbeb8b0fff2a9199f722e3a621a"
Cache-Control: max-age=0""".replace(
    "\n", "\r\n"
)

RAW_HEADERS_WITH_CONNECT = """HTTP/1.1 200 Connection established

HTTP/2 200 
date: Tue, 28 Sep 2021 13:45:34 GMT
content-type: application/epub+zip
content-length: 3706401
content-disposition: filename=ipython-readthedocs-io-en-stable.epub
x-amz-id-2: 2PO2WHP4qGqkhyC1VbRE2KLN2g4uk38vYzaNJDU/OBSxh4lUtYgERD2FNAOPkKPD1a6rsNBMeKI=
x-amz-request-id: 21E21R71FAY4WQKT
last-modified: Sat, 25 Sep 2021 00:43:37 GMT
etag: "6f512f04591f7667486d044c54708448"
x-served: Nginx-Proxito-Sendfile
x-backend: web-i-078619706c1392c2c
x-rtd-project: ipython
x-rtd-version: stable
x-rtd-path: /proxito/epub/ipython/stable/ipython.epub
x-rtd-domain: ipython.readthedocs.io
x-rtd-version-method: path
x-rtd-project-method: subdomain
referrer-policy: no-referrer-when-downgrade
permissions-policy: interest-cohort=()
strict-transport-security: max-age=31536000; includeSubDomains; preload
cf-cache-status: HIT
age: 270
expires: Tue, 28 Sep 2021 15:45:34 GMT
cache-control: public, max-age=7200
accept-ranges: bytes
expect-ct: max-age=604800, report-uri="https://report-uri.cloudflare.com/cdn-cgi/beacon/expect-ct"
server: cloudflare
cf-ray: 695d69b549330686-LHR""".replace(
    "\n", "\r\n"
)


class MyKissHeadersFromStringTest(unittest.TestCase):
    headers: Headers

    def setUp(self) -> None:
        MyKissHeadersFromStringTest.headers = parse_it(RAW_HEADERS)

    def test_decode_partials(self):
        self.assertEqual(
            [("Subject", "pöstal")],
            decode_partials([("Subject", "=?iso-8859-1?q?p=F6stal?=")]),
        )

    def test_bytes_headers(self):
        self.assertEqual(
            MyKissHeadersFromStringTest.headers, parse_it(RAW_HEADERS.encode("utf-8"))
        )

    def test_two_headers_eq(self):
        self.assertEqual(MyKissHeadersFromStringTest.headers, parse_it(RAW_HEADERS))

        self.assertNotEqual(
            MyKissHeadersFromStringTest.headers, parse_it(RAW_HEADERS_MOZILLA)
        )

    def test_headers_get_has(self):
        self.assertIsNone(MyKissHeadersFromStringTest.headers.get("received"))
        self.assertFalse(MyKissHeadersFromStringTest.headers.has("received"))

        self.assertEqual(
            "SAMEORIGIN", MyKissHeadersFromStringTest.headers.get("x-frame-options")
        )

    def test_repr_dict(self):
        dict_ = MyKissHeadersFromStringTest.headers.to_dict()

        self.assertIn("set-cookie", dict_)

        self.assertIn("p3p", dict_)

        self.assertTrue(
            dict_["set-cookie"].startswith(
                "1P_JAR=2020-03-16-21; expires=Wed, 15-Apr-2020 21:27:31 GMT; path=/;"
            )
        )

        self.assertTrue(
            dict_["set-cookie"].endswith(
                "CONSENT=WP.284b10; expires=Fri, 01-Jan-2038 00:00:00 GMT; path=/; domain=.google.fr"
            )
        )

    def test_repr_str(self):
        self.assertEqual(RAW_HEADERS, repr(MyKissHeadersFromStringTest.headers))

        self.assertEqual(RAW_HEADERS, str(MyKissHeadersFromStringTest.headers))

        self.assertEqual(
            "SAMEORIGIN", str(MyKissHeadersFromStringTest.headers.x_frame_options)
        )

        self.assertEqual(
            "x-frame-options: SAMEORIGIN",
            repr(MyKissHeadersFromStringTest.headers.x_frame_options),
        )

    def test_control_basis_exist(self):
        self.assertEqual("DPR", MyKissHeadersFromStringTest.headers.accept_ch)

        self.assertEqual(3, len(MyKissHeadersFromStringTest.headers.set_cookie))

        self.assertIn("Secure", MyKissHeadersFromStringTest.headers.set_cookie[0])

        self.assertEqual(
            "This is not a P3P policy! See g.co/p3phelp for more info.",
            MyKissHeadersFromStringTest.headers.p3p.cp,
        )

        self.assertTrue(MyKissHeadersFromStringTest.headers.has("Cache-Control"))

        self.assertTrue(MyKissHeadersFromStringTest.headers.content_type.has("charset"))

        self.assertEqual(
            "UTF-8", MyKissHeadersFromStringTest.headers.content_type.get("charset")
        )

    def test_control_first_line_not_header(self):
        headers = parse_it(RAW_HEADERS_MOZILLA)

        self.assertEqual(17, len(headers))

        self.assertIn("host", headers)

        self.assertIn("Cache-Control", headers)

    def test_headers_to_bytes(self):
        headers = parse_it(RAW_HEADERS_MOZILLA)

        self.assertEqual(headers, parse_it(bytes(headers)))

    def test_verify_autocompletion_capability(self):
        headers = parse_it(RAW_HEADERS_MOZILLA)

        self.assertIn("accept_encoding", dir(headers))

        self.assertIn("accept_language", dir(headers))

        self.assertTrue(headers.accept)

        self.assertIn("q", dir(headers.accept[-1]))

    def test_fixed_type_output(self):
        headers = parse_it(RAW_HEADERS_MOZILLA)

        self.assertEqual(Header, type(headers.host))

        lock_output_type()

        self.assertEqual(list, type(headers.host))

        self.assertEqual(1, len(headers.host))

        self.assertEqual(list, type(headers.accept[-1].q))

        lock_output_type(False)

        self.assertEqual(Header, type(headers.host))

        self.assertEqual(str, type(headers.accept[-1].q))

    def test_parse_with_extra_connect(self):
        headers: Headers = parse_it(RAW_HEADERS_WITH_CONNECT)

        self.assertTrue("Date" in headers)
        self.assertTrue("Server" in headers)


if __name__ == "__main__":
    unittest.main()



import unittest

from kiss_headers import parse_it


class UnknownMappingTest(unittest.TestCase):
    def test_parse_with_bytes(self):
        headers = parse_it({"User-Agent": b"Hello!"})

        self.assertTrue("User-Agent" in headers)

    def test_parse_with_None(self):
        headers = parse_it({"User-Agent": None, "Test": "Hello!"})

        self.assertTrue("User-Agent" not in headers)
        self.assertTrue("Test" in headers)

    def test_with_type_madness(self):
        headers = parse_it(
            {
                "User-Agent": b"Hello!",
                "Age": 30,
                "Type": 1.554,
                "BeCrazy": {"pff!": "?"},
                "Again": ["a", 0, 8, -1],
            }
        )

        self.assertTrue("User-Agent" in headers)
        self.assertTrue(headers.get("User-agent").content == "Hello!")
        self.assertTrue(headers.get("Age").content == "30")
        self.assertTrue(headers.get("Again").content == '["a", 0, 8, -1]')



import unittest
from datetime import datetime, timezone
from email import utils

from kiss_headers import (
    Allow,
    ContentDisposition,
    ContentLength,
    ContentType,
    CrossOriginResourcePolicy,
    CustomHeader,
    Date,
    From,
    ReferrerPolicy,
    SetCookie,
)


class MyBuilderTestCase(unittest.TestCase):
    def test_custom_header_expect(self):
        with self.assertRaises(NotImplementedError):
            k = CustomHeader("Should absolutely not work !")

    def test_content_type(self):
        self.assertEqual(
            repr(ContentType("application/json", charset="utf-8")),
            'Content-Type: application/json; charset="UTF-8"',
        )

    def test_set_cookie(self):
        dt = datetime.now()

        self.assertEqual(
            repr(SetCookie("MACHINE_IDENTIFIANT", "ABCDEFGHI", expires=dt)),
            'Set-Cookie: MACHINE_IDENTIFIANT="ABCDEFGHI"; expires="{dt}"; HttpOnly'.format(
                dt=utils.format_datetime(dt.astimezone(timezone.utc), usegmt=True)
            ),
        )

    def test_content_length(self):
        self.assertEqual(repr(ContentLength(1881)), "Content-Length: 1881")

    def test_content_disposition(self):
        self.assertEqual(
            repr(ContentDisposition("attachment", filename="test-file.json")),
            'Content-Disposition: attachment; filename="test-file.json"',
        )

    def test_value_error_builder(self):
        with self.assertRaises(ValueError):
            CrossOriginResourcePolicy("policy-not-known-yet")

        with self.assertRaises(ValueError):
            ReferrerPolicy("hello-world")

        with self.assertRaises(ValueError):
            From("not-a-valid-email.com")

        with self.assertRaises(ValueError):
            Allow("DOES-NOT-EXIST-HTTP-VERB")

    def test_verify_always_gmt(self):
        self.assertTrue(repr(Date(datetime.now())).endswith("GMT"))


if __name__ == "__main__":
    unittest.main()



import unittest

from kiss_headers import Header, parse_it


class KissHeadersOperationTest(unittest.TestCase):
    def test_isub(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        headers -= "X-My-Testing"

        self.assertEqual(4, len(headers))

        self.assertNotIn("X-My-Testing", headers)

        headers -= "Received"

        self.assertEqual(2, len(headers))

        self.assertNotIn("Received", headers)

        headers -= Header("X-My-Second-Test", "Precisely")

        self.assertEqual(1, len(headers))

        self.assertIn("X-My-Second-Test", headers)

    def test_sub(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        headers_two = headers - "X-My-Testing"

        self.assertEqual(5, len(headers))

        self.assertEqual(4, len(headers_two))

    def test_add(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        headers_ = headers + Header("content-type", "application/json")

        self.assertEqual(6, len(headers_))

        self.assertIn("content-type", headers_)

        self.assertNotIn("content-type", headers)

    def test_remove_by_del(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        del headers["X-My-Second-Test"]

        self.assertEqual(3, len(headers))

        self.assertNotIn("X-My-Second-Test", headers)

    def test_remove_by_del_attr(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        del headers.x_my_second_test

        self.assertEqual(3, len(headers))

        self.assertNotIn("X-My-Second-Test", headers)

    def test_by_setitem(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        headers["content-type"] = "application/json; charset=utf-8"

        self.assertEqual(6, len(headers))

        self.assertEqual("utf-8", headers.content_type.charset)

    def test_by_setattr(self):
        headers = parse_it(
            """X-My-Testing: 1\nX-My-Second-Test: 1\nX-My-Second-Test: Precisely\nReceived: outpost\nReceived: outpost"""
        )

        self.assertEqual(5, len(headers))

        headers.content_type = "application/json; charset=utf-8"

        self.assertEqual(6, len(headers))

        self.assertEqual("utf-8", headers.content_type.charset)


if __name__ == "__main__":
    unittest.main()



"""
Kiss-Headers
~~~~~~~~~~~~~~

Kiss-Headers is a headers, HTTP or IMAP4 _(message, email)_ flavour, utility, written in pure Python, for humans.
Object oriented headers. Keep it sweet and simple.
Basic usage:

   >>> import requests
   >>> from kiss_headers import parse_it
   >>> r = requests.get('https://www.python.org')
   >>> headers = parse_it(r)
   >>> 'charset' in headers.content_type
   True
   >>> headers.content_type.charset
   'utf-8'
   >>> 'text/html' in headers.content_type
   True
   >>> headers.content_type == 'text/html'
   True
   >>> headers -= 'content-type'
   >>> 'Content-Type' in headers
   False

... or from a raw IMAP4 message:

   >>> message = requests.get("https://gist.githubusercontent.com/Ousret/8b84b736c375bb6aa3d389e86b5116ec/raw/21cb2f7af865e401c37d9b053fb6fe1abf63165b/sample-message.eml").content
   >>> headers = parse_it(message)
   >>> 'Sender' in headers
   True

Others methods and usages are available - see the full documentation
at <https://github.com/Ousret/kiss-headers>.

:copyright: (c) 2020 by Ahmed TAHRI
:license: MIT, see LICENSE for more details.
"""

from .api import dumps, explain, get_polymorphic, parse_it
from .builder import (
    Accept,
    AcceptEncoding,
    AcceptLanguage,
    Allow,
    AltSvc,
    Authorization,
    BasicAuthorization,
    CacheControl,
    Connection,
    ContentDisposition,
    ContentEncoding,
    ContentLength,
    ContentRange,
    ContentSecurityPolicy,
    ContentType,
    CrossOriginResourcePolicy,
    CustomHeader,
    Date,
    Digest,
    Dnt,
    Etag,
    Expires,
    Forwarded,
    From,
    Host,
    IfMatch,
    IfModifiedSince,
    IfNoneMatch,
    IfUnmodifiedSince,
    KeepAlive,
    LastModified,
    Location,
    ProxyAuthorization,
    Referer,
    ReferrerPolicy,
    RetryAfter,
    Server,
    SetCookie,
    StrictTransportSecurity,
    TransferEncoding,
    UpgradeInsecureRequests,
    UserAgent,
    Vary,
    WwwAuthenticate,
    XContentTypeOptions,
    XDnsPrefetchControl,
    XFrameOptions,
    XXssProtection,
)
from .models import Attributes, Header, Headers, lock_output_type
from .serializer import decode, encode
from .version import VERSION, __version__

__all__ = (
    "dumps",
    "explain",
    "get_polymorphic",
    "parse_it",
    "Attributes",
    "Header",
    "Headers",
    "lock_output_type",
    "decode",
    "encode",
    "VERSION",
    "__version__",
    "Accept",
    "AcceptEncoding",
    "AcceptLanguage",
    "Allow",
    "AltSvc",
    "Authorization",
    "BasicAuthorization",
    "CacheControl",
    "Connection",
    "ContentDisposition",
    "ContentEncoding",
    "ContentLength",
    "ContentRange",
    "ContentSecurityPolicy",
    "ContentType",
    "CrossOriginResourcePolicy",
    "CustomHeader",
    "Date",
    "Digest",
    "Dnt",
    "Etag",
    "Expires",
    "Forwarded",
    "From",
    "Host",
    "IfMatch",
    "IfModifiedSince",
    "IfNoneMatch",
    "IfUnmodifiedSince",
    "KeepAlive",
    "LastModified",
    "Location",
    "ProxyAuthorization",
    "Referer",
    "ReferrerPolicy",
    "RetryAfter",
    "Server",
    "SetCookie",
    "StrictTransportSecurity",
    "TransferEncoding",
    "UpgradeInsecureRequests",
    "UserAgent",
    "Vary",
    "WwwAuthenticate",
    "XContentTypeOptions",
    "XDnsPrefetchControl",
    "XFrameOptions",
    "XXssProtection",
)



import unittest

from kiss_headers import Header


class HeaderOrderingTest(unittest.TestCase):
    def test_keep_initial_order(self):
        header = Header("Content-Type", "a; b=k; h; h; z=0")

        self.assertEqual(["a", "b", "h", "h", "z"], header.attrs)

    def test_insertion_in_ordered_header(self):
        header = Header("Content-Type", "a; b=k; h; h; z=0")

        header.insert(2, ppp="nt")

        self.assertEqual(["a", "b", "ppp", "h", "h", "z"], header.attrs)

    def test_pop_in_ordered_header(self):
        header = Header("Content-Type", "a; b=k; h; h; z=0")

        key, value = header.pop(2)

        self.assertEqual(key, "h")

        self.assertIsNone(value)

        self.assertEqual(["a", "b", "h", "z"], header.attrs)

    def test_pop_negative_index(self):
        header = Header("Content-Type", "a; b=k; h; h; z=0")

        key, value = header.pop(-1)

        self.assertEqual(key, "z")

        self.assertEqual(value, "0")

        self.assertEqual(["a", "b", "h", "h"], header.attrs)

    def test_attrs_original_case(self):
        header = Header("Content-Type", "aA; bc=k; hA; h; zZzZ=0")

        with self.subTest(
            "Ensure that attrs and valued_attrs properties keep the original case."
        ):
            self.assertEqual(["aA", "bc", "hA", "h", "zZzZ"], header.attrs)

            self.assertEqual(["bc", "zZzZ"], header.valued_attrs)


if __name__ == "__main__":
    unittest.main()



import unittest

from kiss_headers import Attributes
from kiss_headers.utils import header_content_split


class AttributesTestCase(unittest.TestCase):
    def test_eq(self):
        with self.subTest(
            "Ensure that Attributes instances are compared the correct way"
        ):
            attr_a = Attributes(["a", "p=8a", "a", "XX"])
            attr_b = Attributes(["p=8a", "a", "a", "XX"])
            attr_c = Attributes(["p=8a", "a", "A", "Xx"])
            attr_d = Attributes(["p=8a", "a", "A", "Xx", "XX=a"])
            attr_e = Attributes(["p=8A", "a", "A", "Xx"])

            self.assertEqual(attr_a, attr_b)

            self.assertEqual(attr_a, attr_c)

            self.assertNotEqual(attr_a, attr_d)

            self.assertNotEqual(attr_a, attr_e)

    def test_esc_double_quote(self):
        with self.subTest(
            "Ensure that the double quote character is handled correctly."
        ):
            attributes = Attributes(
                header_content_split(r'text/html; charset="UTF-\"8"', ";")
            )

            self.assertEqual(attributes["charset"], 'UTF-"8')

            self.assertEqual(str(attributes), r'text/html; charset="UTF-\"8"')


if __name__ == "__main__":
    unittest.main()



import unittest

from kiss_headers import (
    AltSvc,
    CacheControl,
    ContentEncoding,
    ContentLength,
    ContentType,
    Date,
    Expires,
    Header,
    Server,
    SetCookie,
    StrictTransportSecurity,
    XFrameOptions,
    XXssProtection,
    explain,
)


class MyExplainTestCase(unittest.TestCase):
    def test_explain_from_objects(self):
        headers = (
            Header("Accept-Ch", "DPR")
            + Header("Accept-Ch-Lifetime", "2592000")
            + AltSvc("quic", ":443", max_age=2592000, versions=["46", "43"])
            + AltSvc("h3-Q050", ":443", max_age=2592000)
            + AltSvc("h3-Q049", ":443", max_age=2592000)
            + AltSvc("h3-Q048", ":443", max_age=2592000)
            + AltSvc("h3-Q046", ":443", max_age=2592000)
            + AltSvc("h3-Q043", ":443", max_age=2592000)
            + CacheControl("private")
            + CacheControl(max_age=0)
            + ContentEncoding("br")
            + ContentLength(64032)
            + ContentType("text/html", charset="utf-8")
            + Date("Mon, 16 Mar 2020 21:27:31 GMT")
            + Expires("-1")
            + Header(
                "P3P", 'CP="This is not a P3P policy! See g.co/p3phelp for more info."'
            )
            + Server("gws")
            + SetCookie(
                "1P_JAR",
                "2020-03-16-21",
                expires="Wed, 15-Apr-2020 21:27:31 GMT",
                path="/",
                domain=".google.fr",
                samesite="none",
                is_secure=True,
                is_httponly=False,
            )
            + SetCookie(
                "NID",
                "200=IGpBMMA3G7tki0niFFATFQ2BnsNceVP6XBtwOutoyw97AJ4_YFT5l1oLfLeX22xeI_STiP4omAB4rmMP3Sxgyo287ldQGwdZSdPOOZ_Md3roDOMAOtXEQ_hFbUvo0VPjS2gL1y00_6kQwpVxCghI2Ozrx-A4Xks3ZIXRj11RsWs",
                expires="Tue, 15-Sep-2020 21:27:31 GMT",
                path="/",
                domain=".google.fr",
                samesite="none",
                is_secure=True,
            )
            + SetCookie(
                "CONSENT",
                "WP.284b10",
                expires="Fri, 01-Jan-2038 00:00:00 GMT",
                path="/",
                domain=".google.fr",
                is_httponly=False,
            )
            + Header("Status", "200")
            + StrictTransportSecurity(max_age=31536000)
            + XFrameOptions("SAMEORIGIN")
            + XXssProtection(False)
        )

        explanations = explain(headers)

        self.assertNotEqual("Unknown explanation.", explanations["Set-Cookie"])

        self.assertNotEqual("Missing docstring.", explanations["Set-Cookie"])

        self.assertEqual("Unknown explanation.", explanations["Accept-Ch"])

        self.assertEqual("Unknown explanation.", explanations["Accept_Ch"])

        self.assertEqual("Unknown explanation.", explanations["aCCept_Ch"])


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from email.message import Message
from email.parser import HeaderParser
from io import BufferedReader, RawIOBase
from json import dumps as json_dumps, loads as json_loads
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Type, TypeVar, Union

from .builder import CustomHeader
from .models import Header, Headers
from .serializer import decode, encode
from .structures import CaseInsensitiveDict
from .utils import (
    class_to_header_name,
    decode_partials,
    extract_class_name,
    extract_encoded_headers,
    header_content_split,
    header_name_to_class,
    is_content_json_object,
    is_legal_header_name,
    normalize_str,
    transform_possible_encoded,
)

T = TypeVar("T", bound=CustomHeader, covariant=True)


def parse_it(raw_headers: Any) -> Headers:
    """
    Just decode anything that could contain headers. That simple PERIOD.
    If passed with a Headers instance, return a deep copy of it.
    :param raw_headers: Accept bytes, str, fp, dict, JSON, email.Message, requests.Response, niquests.Response, urllib3.HTTPResponse and httpx.Response.
    :raises:
        TypeError: If passed argument cannot be parsed to extract headers from it.
    """

    if isinstance(raw_headers, Headers):
        return deepcopy(raw_headers)

    headers: Optional[Iterable[Tuple[Union[str, bytes], Union[str, bytes]]]] = None

    if isinstance(raw_headers, str):
        if raw_headers.startswith("{") and raw_headers.endswith("}"):
            return decode(json_loads(raw_headers))
        headers = HeaderParser().parsestr(raw_headers, headersonly=True).items()
    elif (
        isinstance(raw_headers, bytes)
        or isinstance(raw_headers, RawIOBase)
        or isinstance(raw_headers, BufferedReader)
    ):
        decoded, not_decoded = extract_encoded_headers(
            raw_headers if isinstance(raw_headers, bytes) else raw_headers.read() or b""
        )
        return parse_it(decoded)
    elif isinstance(raw_headers, Mapping) or isinstance(raw_headers, Message):
        headers = raw_headers.items()
    else:
        r = extract_class_name(type(raw_headers))

        if r:
            if r in ["requests.models.Response", "niquests.models.Response"]:
                headers = []
                for header_name in raw_headers.raw.headers:
                    for header_content in raw_headers.raw.headers.getlist(header_name):
                        headers.append((header_name, header_content))
            elif r in [
                "httpx._models.Response",
                "urllib3.response.HTTPResponse",
            ]:  # pragma: no cover
                headers = raw_headers.headers.items()

    if headers is None:
        raise TypeError(  # pragma: no cover
            "Cannot parse type {type_} as it is not supported by kiss-header.".format(
                type_=type(raw_headers)
            )
        )

    revised_headers: List[Tuple[str, str]] = decode_partials(
        transform_possible_encoded(headers)
    )

    # Sometime raw content does not begin with headers. If that is the case, search for the next line.
    if (
        len(revised_headers) == 0
        and len(raw_headers) > 0
        and (isinstance(raw_headers, bytes) or isinstance(raw_headers, str))
    ):
        next_iter = raw_headers.split(
            b"\n" if isinstance(raw_headers, bytes) else "\n", maxsplit=1  # type: ignore
        )

        if len(next_iter) >= 2:
            return parse_it(next_iter[-1])

    # Prepare Header objects
    list_of_headers: List[Header] = []

    for head, content in revised_headers:
        # We should ignore when a illegal name is considered as an header. We avoid ValueError (in __init__ of Header)
        if is_legal_header_name(head) is False:
            continue

        is_json_obj: bool = is_content_json_object(content)
        entries: List[str]

        if is_json_obj is False:
            entries = header_content_split(content, ",")
        else:
            entries = [content]

        # Multiple entries are detected in one content at the only exception that its not IMAP header "Subject".
        if len(entries) > 1 and normalize_str(head) != "subject":
            for entry in entries:
                list_of_headers.append(Header(head, entry))
        else:
            list_of_headers.append(Header(head, content))

    return Headers(*list_of_headers)


def explain(headers: Headers) -> CaseInsensitiveDict:
    """
    Return a brief explanation of each header present in headers if available.
    """
    if not Header.__subclasses__():
        raise LookupError(  # pragma: no cover
            "You cannot use explain() function without properly importing the public package."
        )

    explanations: CaseInsensitiveDict = CaseInsensitiveDict()

    for header in headers:
        if header.name in explanations:
            continue

        try:
            target_class = header_name_to_class(header.name, Header.__subclasses__()[0])
        except TypeError:
            explanations[header.name] = "Unknown explanation."
            continue

        explanations[header.name] = (
            target_class.__doc__.replace("\n", "").lstrip().replace("  ", " ").rstrip()
            if target_class.__doc__
            else "Missing docstring."
        )

    return explanations


def get_polymorphic(
    target: Union[Headers, Header], desired_output: Type[T]
) -> Union[T, List[T], None]:
    """Experimental. Transform a Header or Headers object to its target `CustomHeader` subclass
    to access more ready-to-use methods. eg. You have a Header object named 'Set-Cookie' and you wish
    to extract the expiration date as a datetime.
    >>> header = Header("Set-Cookie", "1P_JAR=2020-03-16-21; expires=Wed, 15-Apr-2020 21:27:31 GMT")
    >>> header["expires"]
    'Wed, 15-Apr-2020 21:27:31 GMT'
    >>> from kiss_headers import SetCookie
    >>> set_cookie = get_polymorphic(header, SetCookie)
    >>> set_cookie.get_expire()
    datetime.datetime(2020, 4, 15, 21, 27, 31, tzinfo=datetime.timezone.utc)
    """

    if not issubclass(desired_output, Header):
        raise TypeError(
            f"The desired output should be a subclass of Header not {desired_output}."
        )

    desired_output_header_name: str = class_to_header_name(desired_output)

    if isinstance(target, Headers):
        r = target.get(desired_output_header_name)

        if r is None:
            return None

    elif isinstance(target, Header):
        if header_name_to_class(target.name, Header) != desired_output:
            raise TypeError(
                f"The target class does not match the desired output class. {target.__class__} != {desired_output}."
            )
        r = target
    else:
        raise TypeError(f"Unable to apply get_polymorphic on type {target.__class__}.")

    # Change __class__ attribute.
    if not isinstance(r, list):
        r.__class__ = desired_output
    else:
        for header in r:
            header.__class__ = desired_output

    return r  # type: ignore


def dumps(headers: Headers, **kwargs: Optional[Any]) -> str:
    return json_dumps(encode(headers), **kwargs)  # type: ignore



from email.header import decode_header
from json import dumps
from re import findall, search, sub
from typing import Any, Iterable, List, Optional, Set, Tuple, Type, Union

RESERVED_KEYWORD: Set[str] = {
    "and_",
    "assert_",
    "in_",
    "not_",
    "pass_",
    "finally_",
    "while_",
    "yield_",
    "is_",
    "as_",
    "break_",
    "return_",
    "elif_",
    "except_",
    "def_",
    "from_",
    "for_",
}


def normalize_str(string: str) -> str:
    """
    Normalize a string by applying on it lowercase and replacing '-' to '_'.
    >>> normalize_str("Content-Type")
    'content_type'
    >>> normalize_str("X-content-type")
    'x_content_type'
    """
    return string.lower().replace("-", "_")


def normalize_list(strings: List[str]) -> List[str]:
    """Normalize a list of string by applying fn normalize_str over each element."""
    return list(map(normalize_str, strings))


def unpack_protected_keyword(name: str) -> str:
    """
    By choice, this project aims to allow developers to access header or attribute in header by using the property
    notation. Some keywords are protected by the language itself. So :
    When starting by a number, prepend an underscore to it. When using a protected keyword, append an underscore to it.
    >>> unpack_protected_keyword("_3to1")
    '3to1'
    >>> unpack_protected_keyword("from_")
    'from'
    >>> unpack_protected_keyword("_from")
    '_from'
    >>> unpack_protected_keyword("3")
    '3'
    >>> unpack_protected_keyword("FroM_")
    'FroM_'
    """
    if len(name) < 2:
        return name

    if name[0] == "_" and name[1].isdigit():
        name = name[1:]

    if name in RESERVED_KEYWORD:
        name = name[:-1]

    return name


def extract_class_name(type_: Type) -> Optional[str]:
    """
    Typically extract a class name from a Type.
    """
    r = findall(r"<class '([a-zA-Z0-9._]+)'>", str(type_))
    return r[0] if r else None


def header_content_split(string: str, delimiter: str) -> List[str]:
    """
    Take a string and split it according to the passed delimiter.
    It will ignore delimiter if inside between double quote, inside a value, or in parenthesis.
    The input string is considered perfectly formed. This function does not split coma on a day
    when attached, see "RFC 7231, section 7.1.1.2: Date".
    >>> header_content_split("Wed, 15-Apr-2020 21:27:31 GMT, Fri, 01-Jan-2038 00:00:00 GMT", ",")
    ['Wed, 15-Apr-2020 21:27:31 GMT', 'Fri, 01-Jan-2038 00:00:00 GMT']
    >>> header_content_split('quic=":443"; ma=2592000; v="46,43", h3-Q050=":443"; ma=2592000, h3-Q049=":443"; ma=2592000', ",")
    ['quic=":443"; ma=2592000; v="46,43"', 'h3-Q050=":443"; ma=2592000', 'h3-Q049=":443"; ma=2592000']
    >>> header_content_split("Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0", ";")
    ['Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0']
    >>> header_content_split("text/html; charset=UTF-8", ";")
    ['text/html', 'charset=UTF-8']
    >>> header_content_split('text/html; charset="UTF-\\\"8"', ";")
    ['text/html', 'charset="UTF-"8"']
    """
    if len(delimiter) != 1 or delimiter not in {";", ",", " "}:
        raise ValueError("Delimiter should be either semi-colon, a coma or a space.")

    in_double_quote: bool = False
    in_parenthesis: bool = False
    in_value: bool = False
    is_on_a_day: bool = False

    result: List[str] = [""]

    for letter, index in zip(string, range(0, len(string))):
        if letter == '"':
            in_double_quote = not in_double_quote

            if in_value and not in_double_quote:
                in_value = False

        elif letter == "(" and not in_parenthesis:
            in_parenthesis = True
        elif letter == ")" and in_parenthesis:
            in_parenthesis = False
        else:
            is_on_a_day = index >= 3 and string[index - 3 : index] in {
                "Mon",
                "Tue",
                "Wed",
                "Thu",
                "Fri",
                "Sat",
                "Sun",
            }

        if not in_double_quote:
            if not in_value and letter == "=":
                in_value = True
            elif letter == ";" and in_value:
                in_value = False

            if in_value and letter == delimiter and not is_on_a_day:
                in_value = False

        if letter == delimiter and (
            (in_value or in_double_quote or in_parenthesis or is_on_a_day) is False
        ):
            result[-1] = result[-1].lstrip().rstrip()
            result.append("")

            continue

        result[-1] += letter

    if result:
        result[-1] = result[-1].lstrip().rstrip()

    return result


def class_to_header_name(type_: Type) -> str:
    """
    Take a type and infer its header name.
    >>> from kiss_headers.builder import ContentType, XContentTypeOptions, BasicAuthorization
    >>> class_to_header_name(ContentType)
    'Content-Type'
    >>> class_to_header_name(XContentTypeOptions)
    'X-Content-Type-Options'
    >>> class_to_header_name(BasicAuthorization)
    'Authorization'
    """
    if hasattr(type_, "__override__") and type_.__override__ is not None:
        return type_.__override__

    class_raw_name: str = str(type_).split("'")[-2].split(".")[-1]

    if class_raw_name.endswith("_"):
        class_raw_name = class_raw_name[:-1]

    if class_raw_name.startswith("_"):
        class_raw_name = class_raw_name[1:]

    header_name: str = str()

    for letter in class_raw_name:
        if letter.isupper() and header_name != "":
            header_name += "-" + letter
            continue
        header_name += letter

    return header_name


def header_name_to_class(name: str, root_type: Type) -> Type:
    """
    The opposite of class_to_header_name function. Will raise TypeError if no corresponding entry is found.
    Do it recursively from the root type.
    >>> from kiss_headers.builder import CustomHeader, ContentType, XContentTypeOptions, LastModified, Date
    >>> header_name_to_class("Content-Type", CustomHeader)
    <class 'kiss_headers.builder.ContentType'>
    >>> header_name_to_class("Last-Modified", CustomHeader)
    <class 'kiss_headers.builder.LastModified'>
    """

    normalized_name = normalize_str(name).replace("_", "")

    for subclass in root_type.__subclasses__():
        class_name = extract_class_name(subclass)

        if class_name is None:
            continue

        if (
            not (
                hasattr(subclass, "__override__") and subclass.__override__ is not None
            )
            and normalize_str(class_name.split(".")[-1]) == normalized_name
        ):
            return subclass

        if subclass.__subclasses__():
            try:
                return header_name_to_class(name, subclass)
            except TypeError:
                continue

    raise TypeError(
        "Cannot find a class matching header named '{name}'.".format(name=name)
    )


def prettify_header_name(name: str) -> str:
    """
    Take a header name and prettify it.
    >>> prettify_header_name("x-hEllo-wORLD")
    'X-Hello-World'
    >>> prettify_header_name("server")
    'Server'
    >>> prettify_header_name("contEnt-TYPE")
    'Content-Type'
    >>> prettify_header_name("content_type")
    'Content-Type'
    """
    return "-".join([el.capitalize() for el in name.replace("_", "-").split("-")])


def decode_partials(items: Iterable[Tuple[str, Any]]) -> List[Tuple[str, str]]:
    """
    This function takes a list of tuples, representing headers by key, value. Where value is bytes or string containing
    (RFC 2047 encoded) partials fragments like the following :
    >>> decode_partials([("Subject", "=?iso-8859-1?q?p=F6stal?=")])
    [('Subject', 'pöstal')]
    """
    revised_items: List[Tuple[str, str]] = list()

    for head, content in items:
        revised_content: str = str()

        for partial, partial_encoding in decode_header(content):
            if isinstance(partial, str):
                revised_content += partial
            if isinstance(partial, bytes):
                revised_content += partial.decode(
                    partial_encoding if partial_encoding is not None else "utf-8",
                    errors="ignore",
                )

        revised_items.append((head, revised_content))

    return revised_items


def unquote(string: str) -> str:
    """
    Remove simple quote or double quote around a string if any.
    >>> unquote('"hello"')
    'hello'
    >>> unquote('"hello')
    '"hello'
    >>> unquote('"a"')
    'a'
    >>> unquote('""')
    ''
    """
    if (
        len(string) >= 2
        and (string.startswith('"') and string.endswith('"'))
        or (string.startswith("'") and string.endswith("'"))
    ):
        return string[1:-1]

    return string


def quote(string: str) -> str:
    """
    Surround string by a double quote char.
    >>> quote("hello")
    '"hello"'
    >>> quote('"hello')
    '""hello"'
    >>> quote('"hello"')
    '"hello"'
    """
    return '"' + unquote(string) + '"'


def count_leftover_space(content: str) -> int:
    """
    A recursive function that counts trailing white space at the end of the given string.
    >>> count_leftover_space("hello   ")
    3
    >>> count_leftover_space("byebye ")
    1
    >>> count_leftover_space("  hello ")
    1
    >>> count_leftover_space("  hello    ")
    4
    """
    if content.endswith(" "):
        return count_leftover_space(content[:-1]) + 1
    return 0


def header_strip(content: str, elem: str) -> str:
    """
    Remove a member for a given header content and take care of the unneeded leftover semi-colon.
    >>> header_strip("text/html; charset=UTF-8; format=flowed", "charset=UTF-8")
    'text/html; format=flowed'
    >>> header_strip("text/html; charset=UTF-8;    format=flowed", "charset=UTF-8")
    'text/html; format=flowed'
    """
    next_semi_colon_index: Optional[int] = None

    try:
        elem_index: int = content.index(elem)
    except ValueError:
        # If the target element in not found within the content, just return the unmodified content.
        return content

    elem_end_index: int = elem_index + len(elem)

    elem = (" " * count_leftover_space(content[:elem_index])) + elem

    try:
        next_semi_colon_index = elem_end_index + content[elem_end_index:].index(";")
    except ValueError:
        pass

    content = (
        content.replace(
            elem
            + (
                content[elem_end_index:next_semi_colon_index] + ";"
                if next_semi_colon_index is not None
                else ""
            ),
            "",
        )
        .rstrip(" ")
        .lstrip(" ")
    )

    if content.startswith(";"):
        content = content[1:]

    if content.endswith(";"):
        content = content[:-1]

    return content


def is_legal_header_name(name: str) -> bool:
    """
    Verify if a provided header name is valid.
    >>> is_legal_header_name(":hello")
    False
    >>> is_legal_header_name("hello")
    True
    >>> is_legal_header_name("Content-Type")
    True
    >>> is_legal_header_name("Hello;")
    False
    >>> is_legal_header_name("Hello\\rWorld")
    False
    >>> is_legal_header_name("Hello \\tWorld")
    False
    >>> is_legal_header_name('Hello World"')
    False
    >>> is_legal_header_name("Hello-World/")
    True
    >>> is_legal_header_name("\x07")
    False
    """
    return (
        name != ""
        and search(r"[^\x21-\x7F]|[:;(),<>=@?\[\]\r\n\t &{}\"\\]", name) is None
    )


def extract_comments(content: str) -> List[str]:
    """
    Extract parts of content that are considered as comments. Between parenthesis.
    >>> extract_comments("Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0 (hello) llll (abc)")
    ['Macintosh; Intel Mac OS X 10.9; rv:50.0', 'hello', 'abc']
    """
    return findall(r"\(([^)]+)\)", content)


def unfold(content: str) -> str:
    r"""Some header content may have folded content (CRLF + n spaces) in it, making your job at reading them a little more difficult.
    This function undoes the folding in the given content.
    >>> unfold("___utmvbtouVBFmB=gZg\r\n    XbNOjalT: Lte; path=/; Max-Age=900")
    '___utmvbtouVBFmB=gZg XbNOjalT: Lte; path=/; Max-Age=900'
    """
    return sub(r"\r\n[ ]+", " ", content)


def extract_encoded_headers(payload: bytes) -> Tuple[str, bytes]:
    """This function's purpose is to extract lines that can be decoded using the UTF-8 decoder.
    >>> extract_encoded_headers("Host: developer.mozilla.org\\r\\nX-Hello-World: 死の漢字\\r\\n\\r\\n".encode("utf-8"))
    ('Host: developer.mozilla.org\\r\\nX-Hello-World: 死の漢字\\r\\n', b'')
    >>> extract_encoded_headers("Host: developer.mozilla.org\\r\\nX-Hello-World: 死の漢字\\r\\n\\r\\nThat IS totally random.".encode("utf-8"))
    ('Host: developer.mozilla.org\\r\\nX-Hello-World: 死の漢字\\r\\n', b'That IS totally random.')
    """
    result: str = ""
    lines: List[bytes] = payload.splitlines()
    index: int = 0

    for line, index in zip(lines, range(0, len(lines))):
        if line == b"":
            return result, b"\r\n".join(lines[index + 1 :])

        try:
            result += line.decode("utf-8") + "\r\n"
        except UnicodeDecodeError:
            break

    return result, b"\r\n".join(lines[index + 1 :])


def unescape_double_quote(content: str) -> str:
    """
    Replace escaped double quote in content by removing the backslash.
    >>> unescape_double_quote(r'UTF\"-8')
    'UTF"-8'
    >>> unescape_double_quote(r'UTF"-8')
    'UTF"-8'
    """
    return content.replace(r"\"", '"')


def escape_double_quote(content: str) -> str:
    r"""
    Replace not escaped double quote in content by adding a backslash beforehand.
    >>> escape_double_quote(r'UTF\"-8')
    'UTF\\"-8'
    >>> escape_double_quote(r'UTF"-8')
    'UTF\\"-8'
    """
    return unescape_double_quote(content).replace('"', r"\"")


def is_content_json_object(content: str) -> bool:
    """
    Sometime, you may receive a header that hold a JSON list or object.
    This function detect it.
    """
    content = content.strip()
    return (content.startswith("{") and content.endswith("}")) or (
        content.startswith("[") and content.endswith("]")
    )


def transform_possible_encoded(
    headers: Iterable[Tuple[Union[str, bytes], Union[str, bytes]]]
) -> Iterable[Tuple[str, str]]:
    decoded = []

    for k, v in headers:
        # we shall discard it if set to None.
        if v is None:
            continue
        if isinstance(k, bytes):
            k = k.decode("utf_8")
        if isinstance(v, bytes):
            v = v.decode("utf_8")
        elif isinstance(v, str) is False:
            if isinstance(v, (dict, list)):
                v = dumps(v)
            else:
                v = str(v)
        decoded.append((k, v))

    return decoded



import unittest
from typing import Optional

from requests import Response, get

from kiss_headers import Authorization, Headers, parse_it


class MyHttpTestKissHeaders(unittest.TestCase):
    HTTPBIN_GET: Optional[Response] = None

    def setUp(self) -> None:
        MyHttpTestKissHeaders.HTTPBIN_GET = get("https://httpbin.org/get")

    def test_httpbin_raw_headers(self):
        headers = parse_it(
            """Host: developer.mozilla.org
User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:50.0) Gecko/20100101 Firefox/50.0
Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
Accept-Language: en-US,en;q=0.5
Accept-Encoding: gzip, deflate, br
Referer: https://developer.mozilla.org/testpage.html
Connection: keep-alive
Upgrade-Insecure-Requests: 1
If-Modified-Since: Mon, 18 Jul 2016 02:36:04 GMT
If-None-Match: "c561c68d0ba92bbeb8b0fff2a9199f722e3a621a"
Cache-Control: max-age=0""".encode(
                "utf-8"
            )
        )

        self.assertEqual(17, len(headers))

        self.assertEqual(
            "c561c68d0ba92bbeb8b0fff2a9199f722e3a621a", headers.if_none_match
        )

        self.assertIn("text/html", headers.accept[0])

        self.assertNotIn("text/htm", headers.accept[0])

        self.assertIn("q", headers.accept[-1])

        self.assertEqual("0", headers.cache_control.max_age)

    def test_parse_response(self):
        headers = parse_it(MyHttpTestKissHeaders.HTTPBIN_GET)

        self.assertEqual(headers.content_type, "application/json")

        self.assertIn("application/json", headers.content_type)

        self.assertNotIn("charset", headers.content_type)

        with self.assertRaises(AttributeError):
            headers.user_agent

    def test_httpbin_get(self):
        response_headers = MyHttpTestKissHeaders.HTTPBIN_GET.headers
        headers = parse_it(response_headers)

        self.assertEqual(headers.content_type, "application/json")

        self.assertIn("application/json", headers.content_type)

        self.assertNotIn("charset", headers.content_type)

        with self.assertRaises(AttributeError):
            headers.user_agent

    def test_httpbin_freeform(self):
        response_headers = get(
            "https://httpbin.org/response-headers",
            params={
                "freeform": 'application/kiss; format="flowed"; expires=Thu, 12 Mar 2020 03:18:25 -0700 (PDT)'
            },
        ).headers

        headers = parse_it(response_headers)

        self.assertIn("freeform", headers)

        self.assertEqual(
            {
                "application/kiss": None,
                "format": "flowed",
                "expires": "Thu, 12 Mar 2020 03:18:25 -0700 (PDT)",
            },
            dict(headers.freeform),
        )

        self.assertIn("application/kiss", headers.freeform)

    def test_httpbin_with_our_headers(self):
        response = get(
            "https://httpbin.org/bearer",
            headers=Headers(Authorization("Bearer", "qwerty")),
        )

        self.assertEqual("qwerty", response.json()["token"])


if __name__ == "__main__":
    unittest.main()



import unittest

from requests import Response, get

from kiss_headers import Headers, decode, dumps, encode, parse_it

RAW_HEADERS: str = """accept-ch: DPR
accept-ch-lifetime: 2592000
alt-svc: quic=":443"; ma=2592000; v="46,43", h3-Q050=":443"; ma=2592000, h3-Q049=":443"; ma=2592000, h3-Q048=":443"; ma=2592000, h3-Q046=":443"; ma=2592000, h3-Q043=":443"; ma=2592000
cache-control: private, max-age=0
content-encoding: br
content-length: 64032
content-type: text/html; charset=UTF-8
date: Mon, 16 Mar 2020 21:27:31 GMT
expires: -1
p3p: CP="This is not a P3P policy! See g.co/p3phelp for more info."
server: gws
set-cookie: 1P_JAR=2020-03-16-21; expires=Wed, 15-Apr-2020 21:27:31 GMT; path=/; domain=.google.fr; Secure; SameSite=none
set-cookie: NID=200=IGpBMMA3G7tki0niFFATFQ2BnsNceVP6XBtwOutoyw97AJ4_YFT5l1oLfLeX22xeI_STiP4omAB4rmMP3Sxgyo287ldQGwdZSdPOOZ_Md3roDOMAOtXEQ_hFbUvo0VPjS2gL1y00_6kQwpVxCghI2Ozrx-A4Xks3ZIXRj11RsWs; expires=Tue, 15-Sep-2020 21:27:31 GMT; path=/; domain=.google.fr; Secure; HttpOnly; SameSite=none
set-cookie: CONSENT=WP.284b10; expires=Fri, 01-Jan-2038 00:00:00 GMT; path=/; domain=.google.fr
status: 200
strict-transport-security: max-age=31536000
thanos: gem=power; gem=mind; gem=soul; gem=space; gem=time; gems; gem
the-one-ring: One ring to rule them all, one ring to find them, One ring to bring them all and in the darkness bind them
x-frame-options: SAMEORIGIN
x-xss-protection: 0""".replace(
    "\n", "\r\n"
)


class SerializerTest(unittest.TestCase):
    def test_encode_decode(self):
        response: Response = get("https://www.httpbin.org/get")
        headers: Headers = parse_it(response)

        encoded_headers = encode(headers)
        decoded_headers = decode(encoded_headers)

        self.assertEqual(
            headers,
            decoded_headers,
            msg="Headers --> fn encode() --> fn decode() --> should be equal to initial object",
        )

    def test_decode_failure(self):
        with self.assertRaises(ValueError):
            decode({"Not Decodable": "X"})  # type: ignore

    def test_json_with_indent(self):
        headers: Headers = parse_it(RAW_HEADERS)
        json_headers: str = dumps(headers, indent=4)

        decoded_headers: Headers = parse_it(json_headers)

        self.assertEqual(
            headers,
            decoded_headers,
            msg="Headers --> json encode --> parse_it --> should be equal",
        )

    def test_json_without_indent(self):
        headers: Headers = parse_it(RAW_HEADERS)
        json_headers: str = dumps(headers)

        decoded_headers: Headers = parse_it(json_headers)

        self.assertEqual(
            headers,
            decoded_headers,
            msg="Headers --> json encode --> parse_it --> should be equal",
        )


if __name__ == "__main__":
    unittest.main()



from base64 import b64decode, b64encode
from datetime import datetime, timezone
from email import utils
from re import findall, fullmatch
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import quote as url_quote, unquote as url_unquote

from .models import Header
from .utils import (
    class_to_header_name,
    header_content_split,
    prettify_header_name,
    quote,
    unquote,
)

"""
Use https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/ to create subclasses of CustomHeader.
"""


class CustomHeader(Header):
    """
    This class is a helper to create ready-to-use Header object with creation assistance.
    Should NOT be instantiated.
    Use this class as a direct parent for creating ready-to-use header object. Inspire yourself with already defined
    class bellow this one.
    """

    __squash__: bool = False  # This value indicate whenever the representation of multiple entries should be squashed into one content.
    __tags__: List[str] = []

    __override__: Optional[
        str
    ] = None  # Create this static member in your custom header when the class name does not match the target header.

    def __init__(self, initial_content: str = "", **kwargs: Optional[str]):
        """
        :param initial_content: Initial content of the Header if any.
        :param kwargs: Provided args. Any key that associate a None value are just ignored.
        """
        if self.__class__ == CustomHeader:
            raise NotImplementedError(
                "You can not instantiate CustomHeader class. You may create first your class that inherit it."
            )

        super().__init__(
            class_to_header_name(self.__class__)
            if not self.__class__.__override__
            else prettify_header_name(self.__class__.__override__),
            initial_content,
        )

        for attribute, value in kwargs.items():
            if value is None:
                continue
            self[attribute] = value


class ContentSecurityPolicy(CustomHeader):
    """
    Content-Security-Policy is the name of a HTTP response header
    that modern browsers use to enhance the security of the document (or web page).
    The Content-Security-Policy header allows you to restrict how resources such as
    JavaScript, CSS, or pretty much anything that the browser loads.
    """

    __tags__ = ["response"]

    def __init__(self, *policies: List[str]):
        """
        :param policies: One policy consist of a list of str like ["default-src", "'none'"].
        >>> header = ContentSecurityPolicy(["default-src", "'none'"], ["img-src", "'self'", "img.example.com"])
        >>> repr(header)
        "Content-Security-Policy: default-src 'none'; img-src 'self' img.example.com"
        >>> header.get_policies_names()
        ['default-src', 'img-src']
        >>> header.get_policy_args("img-src")
        ["'self'", 'img.example.com']
        """
        super().__init__("")

        for policy in policies:
            if len(policy) == 0 or policy[0] not in {
                "default-src",
                "script-src",
                "style-src",
                "img-src",
                "connect-src",
                "font-src",
                "object-src",
                "media-src",
                "frame-src",
                "sandbox",
                "report-uri",
                "child-src",
                "form-action",
                "frame-ancestors",
                "plugin-types",
                "base-uri",
                "report-to",
                "worker-src",
                "manifest-src",
                "prefetch-src",
                "navigate-to",
            }:
                raise ValueError(  # pragma: no cover
                    f"Policy {policy[0]} is not a valid one. See https://content-security-policy.com/ for instructions."
                )
            elif len(policy) == 1:
                raise ValueError(  # pragma: no cover
                    f"Policy {policy[0]} need at least one argument to proceed."
                )

            self += " ".join(policy)  # type: ignore

    def get_policies_names(self) -> List[str]:
        """Fetch a list of policy name set in content."""
        return [member.split(" ")[0] for member in self.attrs]

    def get_policy_args(self, policy_name: str) -> Optional[List[str]]:
        """Retrieve given arguments for a policy."""
        policy_name = policy_name.lower()

        for member in self.attrs:
            parts: List[str] = member.split(" ")

            if parts[0].lower() == policy_name:
                return parts[1:]

        return None


class Accept(CustomHeader):
    """
    The Accept request HTTP header advertises which content types, expressed as MIME types,
    the client is able to understand. Using content negotiation, the server then selects one of
    the proposals, uses it and informs the client of its choice with the Content-Type response header.
    """

    __squash__: bool = True
    __tags__: List[str] = ["request"]

    def __init__(
        self,
        mime: str = "*/*",
        qualifier: float = 1.0,
        **kwargs: Optional[str],
    ):
        """
        :param mime: Describe the MIME using this syntax <MIME_type/MIME_subtype>
        :param qualifier: Any value used is placed in an order of preference expressed using relative quality value called the weight.
        :param kwargs:
        >>> header = Accept("text/html", qualifier=0.8)
        >>> header.content
        'text/html; q="0.8"'
        >>> repr(header)
        'Accept: text/html; q="0.8"'
        >>> header.get_qualifier()
        0.8
        >>> header.get_mime()
        'text/html'
        """
        if len(mime.split("/")) != 2:
            raise ValueError(  # pragma: no cover
                f"The MIME should be described using this syntax <MIME_type/MIME_subtype> not '{mime}'"
            )

        args: Dict = {"q": qualifier if qualifier < 1.0 else None}

        args.update(kwargs)

        super().__init__(
            mime,
            **args,
        )

    def get_mime(self) -> Optional[str]:
        """Return defined mime in current accept header."""
        for el in self.attrs:
            if "/" in el:
                return el
        return None

    def get_qualifier(self, _default: Optional[float] = 1.0) -> Optional[float]:
        """Return defined qualifier for specified mime. If not set, output 1.0."""
        return float(str(self["q"])) if self.has("q") else _default


class ContentType(CustomHeader):
    """
    The Content-Type entity header is used to indicate the media type of the resource.

    In responses, a Content-Type header tells the client what the content type of the returned content actually is.
    Browsers will do MIME sniffing in some cases and will not necessarily follow the value of this header;
    to prevent this behavior, the header X-Content-Type-Options can be set to nosniff.
    """

    __tags__: List[str] = ["request", "response"]

    def __init__(
        self,
        mime: str,
        charset: Optional[str] = None,
        format_: Optional[str] = None,
        boundary: Optional[str] = None,
        **kwargs: Optional[str],
    ):
        """
        :param mime_type: The MIME type of the resource or the data. Format <MIME_type>/<MIME_subtype>.
        :param charset: The character encoding standard. Should be an IANA name.
        :param format_: Mostly used in IMAP, could be one of : original or flowed.
        :param boundary: For multipart entities the boundary directive is required, which consists of 1 to 70 characters from a set of characters known to be very robust through email gateways, and not ending with white space. It is used to encapsulate the boundaries of the multiple parts of the message.
        :param kwargs:
        >>> header = ContentType("text/html", charset="utf-8")
        >>> repr(header)
        'Content-Type: text/html; charset="UTF-8"'
        >>> header.get_charset()
        'UTF-8'
        >>> header.get_mime()
        'text/html'
        """

        if len(mime.split("/")) != 2:
            raise ValueError(  # pragma: no cover
                f"The MIME should be described using this syntax <MIME_type/MIME_subtype> not '{mime}'"
            )

        args: Dict = {
            "charset": charset.upper() if charset else None,
            "format": format_,
            "boundary": boundary,
        }

        args.update(kwargs)

        super().__init__(mime, **args)

    def get_mime(self) -> Optional[str]:
        """Return defined mime in content type."""
        for el in self.attrs:
            if "/" in el:
                return el
        return None

    def get_charset(self, _default: Optional[str] = "ISO-8859-1") -> Optional[str]:
        """Extract defined charset, if not present will return 'ISO-8859-1' by default."""
        return str(self["charset"]) if self.has("charset") else _default


class XContentTypeOptions(CustomHeader):
    """
    The X-Content-Type-Options response HTTP header is a marker used by the server to indicate that
    the MIME types advertised in the Content-Type headers should not be changed and be followed.
    This allows to opt-out of MIME type sniffing, or, in other words, it is a way to say that
    the webmasters knew what they were doing.
    """

    __tags__: List[str] = ["response"]

    def __init__(self, nosniff: bool = True, **kwargs: Optional[str]):
        """
        :param nosniff: see https://fetch.spec.whatwg.org/#x-content-type-options-header
        :param kwargs:
        >>> header = XContentTypeOptions(nosniff=True)
        >>> repr(header)
        'X-Content-Type-Options: nosniff'
        """
        super().__init__("nosniff" if nosniff else "", **kwargs)


class ContentDisposition(CustomHeader):
    """
    In a regular HTTP response, the Content-Disposition response header is a header indicating
    if the content is expected to be displayed inline in the browser, that is, as a Web page or
    as part of a Web page, or as an attachment, that is downloaded and saved locally.
    """

    __tags__: List[str] = ["request", "response"]

    def __init__(
        self,
        disposition: str = "inline",
        name: Optional[str] = None,
        filename: Optional[str] = None,
        fallback_filename: Optional[str] = None,
        boundary: Optional[str] = None,
        **kwargs: Optional[str],
    ):
        """
        :param disposition: Could be either inline, form-data, attachment or empty. Choose one. Default to inline.
        :param name: Is a string containing the name of the HTML field in the form that the content of this subpart refers to.
        :param filename: Is a string containing the original name of the file transmitted. The filename is always optional and must not be used blindly by the application. ASCII-US Only.
        :param fallback_filename: Fallback filename if filename parameter does not uses the encoding defined in RFC 5987. Will be UTF-8 (url-quote) encoded.
        :param boundary: For multipart entities the boundary directive is required, which consists of 1 to 70 characters from a set of characters known to be very robust through email gateways, and not ending with white space. It is used to encapsulate the boundaries of the multiple parts of the message.
        :param kwargs:
        >>> header = ContentDisposition("attachment", filename="hello-world.pdf", fallback_filename="こんにちは世界.pdf")
        >>> repr(header)
        'Content-Disposition: attachment; filename="hello-world.pdf"; filename*="UTF-8\\'\\'%E3%81%93%E3%82%93%E3%81%AB%E3%81%A1%E3%81%AF%E4%B8%96%E7%95%8C.pdf"'
        >>> header.get_disposition()
        'attachment'
        >>> header.get_filename_decoded()
        'こんにちは世界.pdf'
        """
        if disposition not in ["attachment", "inline", "form-data", ""]:
            raise ValueError(  # pragma: no cover
                "Disposition should be either inline, form-data, attachment or empty. Choose one."
            )

        if filename:
            try:
                filename.encode("ASCII")
            except UnicodeEncodeError:
                raise ValueError(  # pragma: no cover
                    "The filename should only contain valid ASCII characters. Not '{fb_filename}'. Use fallback_filename instead.".format(
                        fb_filename=fallback_filename
                    )
                )

        args: Dict = {
            "name": name,
            "filename": filename,
            "filename*": ("UTF-8''" + url_quote(fallback_filename, encoding="utf-8"))
            if fallback_filename
            else None,
            "boundary": boundary,
        }

        args.update(kwargs)

        super().__init__(
            disposition,
            **args,
        )

    def get_disposition(self) -> Optional[str]:
        """Extract set disposition from Content-Disposition"""
        for attr in self.attrs:
            if attr.lower() in ["attachment", "inline", "form-data"]:
                return attr

        return None

    def get_filename_decoded(self) -> Optional[str]:
        """Retrieve and decode if necessary the associated filename."""
        if "filename*" in self:
            try:
                encoding, encoded_filename = tuple(str(self["filename*"]).split("''"))
                return url_unquote(encoded_filename, encoding)
            except ValueError:
                pass

        return str(self["filename"]) if "filename" in self else None


class Authorization(CustomHeader):
    """
    The HTTP Authorization request header contains the credentials to authenticate a user agent with a server,
    usually, but not necessarily, after the server has responded with a 401 Unauthorized status
    and the WWW-Authenticate header.
    """

    __tags__: List[str] = ["request"]

    def __init__(self, type_: str, credentials: str, **kwargs: Optional[str]):
        """
        :param type_: Authentication type. A common type is "Basic". See IANA registry of Authentication schemes for others.
        :param credentials: Associated credentials to use. Preferably Base-64 encoded.
        >>> header = Authorization("Bearer", "base64encoded")
        >>> repr(header)
        'Authorization: Bearer base64encoded'
        """
        if type_.lower() not in [
            "basic",
            "bearer",
            "digest",
            "hoba",
            "mutual",
            "negotiate",
            "oauth",
            "scram-sha-1",
            "scram-sha-256",
            "vapid",
            "aws4-hmac-sha256",
            "ntlm",
        ]:
            raise ValueError(  # pragma: no cover
                "Authorization type should exist in IANA registry of Authentication schemes"
            )

        super().__init__(
            "{type_} {credentials}".format(type_=type_, credentials=credentials),
            **kwargs,
        )

    def get_auth_type(self) -> str:
        """Return the auth type used in Authorization."""
        return self.content.split(" ", maxsplit=1)[0]

    def get_credentials(self) -> str:
        """Output the credentials."""
        return self.content.split(" ", maxsplit=1)[1]


class BasicAuthorization(Authorization):
    """
    Same as Authorization header but simplified for the Basic method. Also an example of __override__ usage.
    """

    __override__ = "Authorization"

    def __init__(
        self,
        username: str,
        password: str,
        charset: str = "latin1",
        **kwargs: Optional[str],
    ):
        """
        :param username:
        :param password:
        :param charset: By default, credentials are encoded using latin1 charset. You may want to choose otherwise.
        :param kwargs:
        >>> header = BasicAuthorization("azerty", "qwerty")
        >>> header
        Authorization: Basic YXplcnR5OnF3ZXJ0eQ==
        >>> header.get_username_password()
        ('azerty', 'qwerty')
        """
        if ":" in username:
            raise ValueError(  # pragma: no cover
                "The username cannot contain a single colon in it."
            )

        b64_auth_content: str = b64encode(
            (username + ":" + password).encode(charset)
        ).decode("ascii")

        super().__init__("Basic", b64_auth_content, **kwargs)

    def get_credentials(self, __default_charset: str = "latin1") -> str:
        """Decode base64 encoded credentials from Authorization header."""
        if self.get_auth_type().lower() != "basic":
            raise ValueError(  # pragma: no cover
                f"Only Authorization using Basic method is supported by BasicAuthorization. Given '{self.get_auth_type()}'."
            )

        return b64decode(super().get_credentials()).decode(__default_charset)

    def get_username_password(
        self, __default_charset: str = "latin1"
    ) -> Tuple[str, ...]:
        """Extract username and password as a tuple from Basic Authorization."""
        return tuple(self.get_credentials(__default_charset).split(":", maxsplit=1))


class ProxyAuthorization(Authorization):
    """
    The HTTP Proxy-Authorization request header contains the credentials to authenticate a user agent to a proxy server,
    usually after the server has responded with a 407 Proxy Authentication Required status
    and the Proxy-Authenticate header.
    """

    __tags__: List[str] = ["request"]

    def __init__(self, type_: str, credentials: str):
        """
        :param type_: Authentication type. A common type is "Basic". See IANA registry of Authentication schemes for others.
        :param credentials: Associated credentials to use. Preferably Base-64 encoded.
        """
        super().__init__(type_, credentials)


class Host(CustomHeader):
    """
    The Host request header specifies the domain name of the server (for virtual hosting),
    and (optionally) the TCP port number on which the server is listening.
    """

    __tags__: List[str] = ["request"]

    def __init__(self, host: str, port: Optional[int] = None, **kwargs: Optional[str]):
        """
        :param host: The domain name of the server (for virtual hosting).
        :param port: TCP port number on which the server is listening.
        >>> header = Host("www.python.org")
        >>> repr(header)
        'Host: www.python.org'
        >>> header = Host("www.python.org", port=8000)
        >>> repr(header)
        'Host: www.python.org:8000'
        """
        super().__init__(host + (":" + str(port) if port else ""), **kwargs)


class Connection(CustomHeader):
    """
    The Connection general header controls whether or not the network connection stays open after the current transaction finishes.
    If the value sent is keep-alive, the connection is persistent and not closed, allowing for subsequent requests to the same server to be done.
    """

    __tags__ = ["request", "response"]

    def __init__(self, should_keep_alive: bool, **kwargs: Optional[str]):
        """
        :param should_keep_alive: Indicates that the client would like to keep the connection open or not.
        :param kwargs:
        """
        super().__init__("keep-alive" if should_keep_alive else "close", **kwargs)


class ContentLength(CustomHeader):
    """
    The Content-Length entity header indicates the size of the entity-body, in bytes, sent to the recipient.
    """

    __tags__: List[str] = ["request", "response"]

    def __init__(self, length: int, **kwargs: Optional[str]):
        """
        :param length: The length in decimal number of octets.
        """
        super().__init__(str(length), **kwargs)


class Date(CustomHeader):
    """
    The Date general HTTP header contains the date and time at which the message was originated.
    """

    __tags__: List[str] = ["response"]

    def __init__(self, my_date: Union[datetime, str], **kwargs: Optional[str]):
        """
        :param my_date: Can either be a datetime that will be automatically converted or a raw string.
        :param kwargs:
        """
        super().__init__(
            utils.format_datetime(my_date.astimezone(timezone.utc), usegmt=True)
            if not isinstance(my_date, str)
            else my_date,
            **kwargs,
        )

    def get_datetime(self) -> datetime:
        """Parse and return a datetime according to content."""
        return utils.parsedate_to_datetime(str(self))


class CrossOriginResourcePolicy(CustomHeader):
    """
    The HTTP Cross-Origin-Resource-Policy response header conveys a desire that
    the browser blocks no-cors cross-origin/cross-site requests to the given resource.
    """

    __tags__: List[str] = ["response"]

    def __init__(self, policy: str, **kwargs: Optional[str]):
        """
        :param policy: Accepted values are same-site, same-origin or cross-origin.
        :param kwargs:
        """
        policy = policy.lower()

        if policy not in ["same-site", "same-origin", "cross-origin"]:
            raise ValueError(  # pragma: no cover
                "'{policy}' is not a recognized policy for Cross-Origin-Resource-Policy. Accepted values are same-site, same-origin or cross-origin."
            )

        super().__init__(policy, **kwargs)


class Allow(CustomHeader):
    """
    The Allow header lists the set of methods supported by a resource.
    """

    __tags__ = ["response"]
    __squash__ = True

    def __init__(self, supported_verb: str, **kwargs: Optional[str]):
        """
        :param supported_verb: Choose exactly one of "HEAD", "GET", "POST", "PUT", "PATCH", "DELETE", "PURGE", "CONNECT" or "TRACE" HTTP verbs.
        :param kwargs:
        >>> header = Allow("POST")
        >>> repr(header)
        'Allow: POST'
        """
        supported_verb = supported_verb.upper()

        if supported_verb not in [
            "HEAD",
            "GET",
            "POST",
            "PUT",
            "PATCH",
            "DELETE",
            "PURGE",
            "CONNECT",
            "TRACE",
        ]:
            raise ValueError(  # pragma: no cover
                "'{verb}' is not a supported verb. Please choose only one HTTP verb per Allow header."
            )

        super().__init__(supported_verb, **kwargs)


class Digest(CustomHeader):
    """
    The Digest response HTTP header provides a digest of the requested resource. RFC 7231.
    """

    __tags__ = ["response"]

    def __init__(self, algorithm: str, value: str, **kwargs: Optional[str]):
        """
        :param algorithm: Supported digest algorithms are defined in RFC 3230 and RFC 5843, and include SHA-256 and SHA-512.
        :param value: The result of applying the digest algorithm to the resource representation and encoding the result.
        :param kwargs:
        """
        args: Dict = {algorithm: value}
        args.update(kwargs)

        super().__init__("", **args)


class Cookie(CustomHeader):
    """The Cookie HTTP request header contains stored HTTP cookies previously sent by
    the server with the Set-Cookie header."""

    __tags__ = ["request"]

    def __init__(self, **kwargs: Optional[str]):
        """
        :param kwargs: Pair of cookie name associated with a value.
        """
        super().__init__("", **kwargs)

    def get_cookies_names(self) -> List[str]:
        """Retrieve all defined cookie names from Cookie header."""
        return self.attrs

    def get_cookie_value(
        self, cookie_name: str, __default: Optional[str] = None
    ) -> Optional[str]:
        """Retrieve associated value with a given cookie name."""
        return (
            str(self[cookie_name]).replace('\\"', "")
            if cookie_name in self
            else __default
        )


class SetCookie(CustomHeader):
    """
    The Set-Cookie HTTP response header is used to send cookies from the server to the user agent,
    so the user agent can send them back to the server later.
    """

    __tags__: List[str] = ["response"]

    def __init__(
        self,
        cookie_name: str,
        cookie_value: str,
        expires: Optional[Union[datetime, str]] = None,
        max_age: Optional[int] = None,
        domain: Optional[str] = None,
        path: Optional[str] = None,
        samesite: Optional[str] = None,
        is_secure: bool = False,
        is_httponly: bool = True,
        **kwargs: Optional[str],
    ):
        """
        :param cookie_name: Can be any US-ASCII characters, except control characters, spaces, or tabs.
        :param cookie_value: Can include any US-ASCII characters excluding control characters, Whitespace, double quotes, comma, semicolon, and backslash.
        :param expires: The maximum lifetime of the cookie as an HTTP-date timestamp. Provided datetime will be converted automatically.
        :param max_age: Number of seconds until the cookie expires. A zero or negative number will expire the cookie immediately. If both Expires and Max-Age are set, Max-Age has precedence.
        :param domain: Hosts to where the cookie will be sent. If omitted, defaults to the host of the current document URL, not including subdomains.
        :param path: A path that must exist in the requested URL, or the browser won't send the Cookie header.
        :param samesite: Asserts that a cookie must not be sent with cross-origin requests, providing some protection against cross-site request forgery attacks.
        :param is_secure: A secure cookie is only sent to the server when a request is made with the https: scheme.
        :param is_httponly: Forbids JavaScript from accessing the cookie.
        :param kwargs:
        """

        for letter in cookie_name:
            if letter in {'<>@,;:\\"/[]?={}'}:
                raise ValueError(  # pragma: no cover
                    'The cookie name can not contains any of the following char: <>@,;:"/[]?={}'
                )

        if samesite and samesite.lower() not in ["strict", "lax", "none"]:
            raise ValueError(  # pragma: no cover
                "Samesite attribute can only be one of the following: Strict, Lax or None."
            )

        args: Dict = {
            cookie_name: cookie_value,
            "expires": utils.format_datetime(
                expires.astimezone(timezone.utc), usegmt=True
            )
            if isinstance(expires, datetime)
            else expires,
            "max-age": max_age,
            "domain": domain,
            "path": path,
            "samesite": samesite,
        }

        args.update(kwargs)

        super().__init__("", **args)

        if is_secure:
            self += "Secure"  # type: ignore
        if is_httponly:
            self += "HttpOnly"  # type: ignore

    def is_http_only(self) -> bool:
        """Determine if the cookie can only be accessed by the browser."""
        return "HttpOnly" in self

    def is_secure(self) -> bool:
        """Determine if the cookie is TLS/SSL only."""
        return "Secure" in self

    def get_expire(self) -> Optional[datetime]:
        """Retrieve the parsed expiration date."""
        return (
            utils.parsedate_to_datetime(str(self["expires"]))
            if self.has("expires")
            else None
        )

    def get_max_age(self) -> Optional[int]:
        """Getting the max-age value as an integer if set."""
        return int(str(self["max-age"])) if "max-age" in self else None

    def get_cookie_name(self) -> str:
        """Extract the cookie name."""
        return self.attrs[0]

    def get_cookie_value(self) -> str:
        """Extract the cookie value."""
        return str(self[self.get_cookie_name()]).replace('\\"', "")


class StrictTransportSecurity(CustomHeader):
    """
    The HTTP Strict-Transport-Security response header (often abbreviated as HSTS) lets a web site
    tell browsers that it should only be accessed using HTTPS, instead of using HTTP.
    """

    __tags__: List[str] = ["response"]

    def __init__(
        self,
        max_age: int,
        does_includesubdomains: bool = False,
        is_preload: bool = False,
        **kwargs: Optional[str],
    ):
        """
        :param max_age: The time, in seconds, that the browser should remember that a site is only to be accessed using HTTPS.
        :param does_includesubdomains: If this optional parameter is specified, this rule applies to all of the site's subdomains as well.
        :param is_preload: Preloading Strict Transport Security. Google maintains an HSTS preload service. By following the guidelines and successfully submitting your domain, browsers will never connect to your domain using an insecure connection.
        :param kwargs:
        """
        args: Dict = {"max-age": max_age}

        args.update(kwargs)

        super().__init__("", **args)

        if does_includesubdomains:
            self += "includeSubDomains"  # type: ignore

        if is_preload:
            self += "preload"  # type: ignore

    def does_includesubdomains(self) -> bool:
        """Verify if this rule applies to all of the site's subdomains."""
        return "includeSubDomains" in self

    def should_preload(self) -> bool:
        """Verify if Preloading Strict Transport Security should be set."""
        return "preload" in self

    def get_max_age(self) -> Optional[int]:
        """Get the time, in seconds, if set, that the browser should remember."""
        return int(str(self["max-age"])) if self.has("max-age") else None


class UpgradeInsecureRequests(CustomHeader):
    """
    The HTTP Upgrade-Insecure-Requests request header sends a signal to the server expressing
    the client’s preference for an encrypted and authenticated response, and that it
    can successfully handle the upgrade-insecure-requests CSP directive.
    """

    __tags__: List[str] = ["request", "response"]

    def __init__(self, **kwargs: Optional[str]):
        super().__init__("1", **kwargs)


class TransferEncoding(CustomHeader):
    """
    The Transfer-Encoding header specifies the form of encoding used to safely transfer the payload body to the user.
    """

    __tags__: List[str] = ["response"]
    __squash__: bool = True

    def __init__(
        self,
        method: str,
        **kwargs: Optional[str],
    ):
        """
        :param method: Either chunked, compress, deflate, gzip, identity or br.
        :param kwargs:
        """

        method = method.lower()

        if method not in [
            "chunked",
            "compress",
            "deflate",
            "gzip",
            "identity",
            "br",
            "*",
        ]:
            raise ValueError(  # pragma: no cover
                "You should choose between 'chunked', 'compress', 'deflate', 'gzip', 'identity' or 'br' for the encoding method."
            )

        super().__init__(method, **kwargs)


class ContentEncoding(TransferEncoding):
    """
    The Content-Encoding entity header is used to compress the media-type. When present,
    its value indicates which encodings were applied to the entity-body. It lets the client
    know how to decode in order to obtain the media-type referenced by the Content-Type header.

    If multiple, keep them in the order in which they were applied.
    """

    __tags__ = ["request", "response"]

    def __init__(self, method: str, **kwargs: Optional[str]):
        """
        :param method: Either chunked, compress, deflate, gzip, identity or br.
        :param kwargs:
        """

        super().__init__(method, **kwargs)


class AcceptEncoding(TransferEncoding):
    """
    The Accept-Encoding request HTTP header advertises which content encoding, usually a compression algorithm,
    the client is able to understand. Using content negotiation, the server selects one of the proposals,
    uses it and informs the client of its choice with the Content-Encoding response header.
    """

    __tags__ = ["request"]

    def __init__(self, method: str, qualifier: float = 1.0, **kwargs: Optional[str]):
        """
        :param method: Either chunked, compress, deflate, gzip, identity, br or a wildcard.
        :param qualifier: Any value used is placed in an order of preference expressed using relative quality value called the weight.
        :param kwargs:
        """
        args: Dict = {"q": qualifier if qualifier != 1.0 else None}

        args.update(kwargs)

        super().__init__(method, **args)

    def get_qualifier(self, _default: Optional[float] = 1.0) -> Optional[float]:
        """Return defined qualifier for specified encoding. If not set, output 1.0."""
        return float(str(self["q"])) if self.has("q") else _default


class Dnt(CustomHeader):
    """
    The DNT (Do Not Track) request header indicates the user's tracking preference.
    It lets users indicate whether they would prefer privacy rather than personalized content.
    """

    __tags__: List[str] = ["request"]

    def __init__(self, tracking_consent: bool = False, **kwargs: Optional[str]):
        """
        :param tracking_consent: The user prefers to allow tracking on the target site or not.
        :param kwargs:
        """
        super().__init__("0" if tracking_consent else "1", **kwargs)


class UserAgent(CustomHeader):
    """
    The User-Agent request header is a characteristic string that lets servers and network
    peers identify the application, operating system, vendor, and/or version of the requesting user agent.
    """

    __tags__: List[str] = ["request"]

    def __init__(self, characteristics: str, **kwargs: Optional[str]):
        super().__init__(characteristics, **kwargs)


class AltSvc(CustomHeader):
    """
    The Alt-Svc HTTP response header is used to advertise alternative services through which
    the same resource can be reached. An alternative service is defined by a protocol/host/port combination.
    """

    __tags__: List[str] = ["response"]
    __squash__: bool = True

    def __init__(
        self,
        protocol_id: str,
        alt_authority: str,
        max_age: Optional[int] = None,
        versions: Optional[List[str]] = None,
        do_persist: Optional[bool] = None,
        **kwargs: Optional[str],
    ):
        """
        :param protocol_id: The ALPN protocol identifier. Examples include h2 for HTTP/2 and h3-25 for draft 25 of the HTTP/3 protocol.
        :param alt_authority: The quoted string specifying the alternative authority which consists of an optional host override, a colon, and a mandatory port number.
        :param max_age: The number of seconds for which the alternative service is considered fresh. If omitted, it defaults to 24 hours.
        :param versions: List of supported versions of the protocol id if the protocol id can be ambiguous. (like QUIC)
        :param do_persist: Use the parameter to ensures that the entry is not deleted through network configuration changes.
        :param kwargs:
        """
        args: Dict = {
            protocol_id: alt_authority,
            "ma": max_age,
            "persist": 1 if do_persist else None,
            "v": ",".join(versions) if versions else None,
        }

        args.update(kwargs)

        super().__init__(**args)

    def get_protocol_id(self) -> str:
        """Get the ALPN protocol identifier."""
        return self.attrs[0]

    def get_alt_authority(self) -> str:
        """Extract the alternative authority which consists of an optional host override, a colon, and a mandatory port number."""
        return str(self[self.get_protocol_id()])

    def get_max_age(self) -> Optional[int]:
        """Output the number of seconds for which the alternative service is considered fresh. None if undefined."""
        return int(str(self["ma"])) if "ma" in self else None

    def get_versions(self) -> Optional[List[str]]:
        """May return, if available, a list of versions of the ALPN protocol identifier."""
        return str(self["v"]).split(",") if "v" in self else None

    def should_persist(self) -> Optional[bool]:
        """Verify if the entry should not be deleted through network configuration changes. None if no indication."""
        return str(self["persist"]) == "1" if "persist" in self else None


class Forwarded(CustomHeader):
    """
    The Forwarded header contains information from the client-facing side of proxy servers
    that is altered or lost when a proxy is involved in the path of the request.
    """

    __tags__: List[str] = ["request", "response"]

    def __init__(
        self,
        by: str,
        for_: str,
        using_proto: str,
        host: Optional[str] = None,
        **kwargs: Optional[str],
    ):
        """
        :param by: The interface where the request came in to the proxy server. Could be an IP address, an obfuscated identifier or "unknown".
        :param for_: The client that initiated the request and subsequent proxies in a chain of proxies.
        :param host: The Host request header field as received by the proxy.
        :param using_proto: Indicates which protocol was used to make the request (typically "http" or "https").
        :param kwargs:
        """
        args: Dict = {"by": by, "for": for_, "host": host, "proto": using_proto}

        args.update(kwargs)

        super().__init__("", **args)


class LastModified(Date):
    """
    The Last-Modified response HTTP header contains the date and time at which the origin server
    believes the resource was last modified. It is used as a validator
    to determine if a resource received or stored is the same.
    """

    __tags__: List[str] = ["response"]

    def __init__(self, my_date: Union[datetime, str], **kwargs: Optional[str]):
        """
        :param my_date:
        :param kwargs:
        """
        super().__init__(my_date, **kwargs)


class Referer(CustomHeader):
    """
    The Referer request header contains the address of the previous web page from which a link to the currently
    requested page was followed. The Referer header allows servers to identify where people are
    visiting them from and may use that data for analytics, logging, or optimized caching, for example.

    Note that referer is actually a misspelling of the word "referrer". See https://en.wikipedia.org/wiki/HTTP_referer
    """

    __tags__: List[str] = ["request"]

    def __init__(self, url: str, **kwargs: Optional[str]):
        """
        :param url: An absolute or partial address of the previous web page from which a link to the currently requested page was followed. URL fragments not included.
        :param kwargs:
        """
        super().__init__(url, **kwargs)


class ReferrerPolicy(CustomHeader):
    """
    The Referrer-Policy HTTP header controls how much referrer information
    (sent via the Referer header) should be included with requests.
    """

    __tags__: List[str] = ["response"]
    __squash__ = True

    def __init__(self, policy: str, **kwargs: Optional[str]):
        """
        :param policy: Either "no-referrer", "no-referrer-when-downgrade", "origin", "origin-when-cross-origin", "same-origin", "strict-origin", "strict-origin-when-cross-origin", "unsafe-url"
        :param kwargs:
        """
        if policy not in [
            "no-referrer",
            "no-referrer-when-downgrade",
            "origin",
            "origin-when-cross-origin",
            "same-origin",
            "strict-origin",
            "strict-origin-when-cross-origin",
            "unsafe-url",
        ]:
            raise ValueError(  # pragma: no cover
                f"'{policy}' is not a valid referrer policy. Please choose only one per ReferrerPolicy instance"
            )

        super().__init__(policy, **kwargs)


class RetryAfter(Date):
    """
    The Retry-After response HTTP header indicates how long the user agent should wait
    before making a follow-up request.
    """

    __tags__: List[str] = ["response"]

    def __init__(self, delay_or_date: Union[datetime, int], **kwargs: Optional[str]):
        super().__init__(
            delay_or_date
            if isinstance(delay_or_date, datetime)
            else str(delay_or_date),
            **kwargs,
        )


class AcceptLanguage(CustomHeader):
    """
    The Accept-Language request HTTP header advertises which languages the client is able to understand,
    and which locale variant is preferred. (By languages, we mean natural languages,
    such as English, and not programming languages.)
    """

    __squash__: bool = True
    __tags__: List[str] = ["request"]

    def __init__(
        self, language: str = "*", qualifier: float = 1.0, **kwargs: Optional[str]
    ):
        """
        :param language: A language tag (which is sometimes referred to as a "locale identifier"). This consists of a 2-3 letter base language tag representing the language.
        :param qualifier: Any value placed in an order of preference expressed using a relative quality value called weight.
        :param kwargs:
        """
        args: Dict = {"q": qualifier if qualifier < 1.0 else None}

        args.update(kwargs)

        super().__init__(
            language,
            **args,
        )

    def get_qualifier(self, _default: Optional[float] = 1.0) -> Optional[float]:
        """Return defined qualifier for specified language. If not set, output 1.0."""
        return float(str(self["q"])) if self.has("q") else _default


class Etag(CustomHeader):
    """
    The ETag HTTP response header is an identifier for a specific version of a resource.
    It lets caches be more efficient and save bandwidth, as a web server does not need to
    resend a full response if the content has not changed.
    """

    __tags__: List[str] = ["response"]

    def __init__(
        self,
        etag_value: str,
        is_a_weak_validator: bool = False,
        **kwargs: Optional[str],
    ):
        """
        :param etag_value: Entity tag uniquely representing the requested resource. ASCII string only. Not quoted.
        :param is_a_weak_validator: Indicates that a weak validator is used. Weak etags are easy to generate, but are far less useful for comparisons.
        :param kwargs:
        """
        super().__init__(
            "{weak_validation_cond}{etag}".format(
                weak_validation_cond="W/" if is_a_weak_validator else "",
                etag=quote(etag_value),
            ),
            **kwargs,
        )


class XFrameOptions(CustomHeader):
    """
    The X-Frame-Options HTTP response header can be used to indicate whether or not a browser
    should be allowed to render a page in a <frame>, <iframe>, <embed> or <object>. Sites can use this to
    avoid clickjacking attacks, by ensuring that their content is not embedded into other sites.
    """

    __tags__: List[str] = ["response"]

    def __init__(self, policy: str, **kwargs: Optional[str]):
        """
        :param policy: Can be either DENY or SAMEORIGIN.
        :param kwargs:
        """
        policy = policy.upper()

        if policy not in ["DENY", "SAMEORIGIN"]:
            raise ValueError(  # pragma: no cover
                "'{policy}' is not a valid X-Frame-Options policy. Choose between DENY and SAMEORIGIN."
            )

        super().__init__(policy, **kwargs)


class XXssProtection(CustomHeader):
    """
    The HTTP X-XSS-Protection response header is a feature of Internet Explorer, Chrome and Safari that
    stops pages from loading when they detect reflected cross-site scripting (XSS) attacks.
    Although these protections are largely unnecessary in modern browsers when sites implement a strong
    Content-Security-Policy that disables the use of inline JavaScript
    """

    __tags__: List[str] = ["response"]

    def __init__(
        self,
        enable_filtering: bool = True,
        enable_block_rendering: bool = False,
        report_uri: Optional[str] = None,
        **kwargs: Optional[str],
    ):
        """
        :param enable_filtering: Enables XSS filtering (usually default in browsers). If a cross-site scripting attack is detected, the browser will sanitize the page (remove the unsafe parts).
        :param enable_block_rendering: Rather than sanitizing the page, the browser will prevent rendering of the page if an attack is detected.
        :param report_uri: (Chromium only) If a cross-site scripting attack is detected, the browser will sanitize the page and report the violation. This uses the functionality of the CSP report-uri directive to send a report.
        :param kwargs:
        """
        if enable_filtering is False:
            super().__init__("0", **kwargs)
            return

        args: Dict = {
            "mode": "block" if enable_block_rendering else None,
            "report": report_uri,
        }

        args.update(kwargs)

        super().__init__("1", **args)


class WwwAuthenticate(CustomHeader):
    """
    The HTTP WWW-Authenticate response header defines the authentication
    method that should be used to gain access to a resource.
    Fair-Warning : This header is like none other and is harder to parse. It need a specific case.
    """

    __tags__: List[str] = ["response"]
    __squash__ = True

    def __init__(
        self,
        auth_type: Optional[str] = None,
        challenge: str = "realm",
        value: str = "Secured area",
        **kwargs: Optional[str],
    ):
        """
        >>> www_authenticate = WwwAuthenticate("Basic", "realm", "Secured area")
        >>> repr(www_authenticate)
        'Www-Authenticate: Basic realm="Secured area"'
        >>> headers = www_authenticate + WwwAuthenticate(challenge="charset", value="UTF-8")
        >>> repr(headers)
        'Www-Authenticate: Basic realm="Secured area", charset="UTF-8"'
        >>> www_authenticate.get_challenge()
        ('realm', 'Secured area')
        >>> www_authenticate.get_auth_type()
        'Basic'
        """
        super().__init__(
            f'{auth_type+" " if auth_type else ""}{challenge}="{value}"', **kwargs
        )

    def get_auth_type(self) -> Optional[str]:
        """Retrieve given authentication method if defined."""
        parts: List[str] = header_content_split(str(self), " ")

        if len(parts) >= 1 and "=" not in parts:
            return parts[0]

        return None

    def get_challenge(self) -> Tuple[str, str]:
        """Output a tuple containing the challenge and the associated value. Raises :ValueError:"""
        parts: List[str] = header_content_split(str(self), " ")

        for part in parts:
            if "=" in part:
                challenge, value = tuple(part.split("=", maxsplit=1))
                return challenge, unquote(value)

        raise ValueError(  # pragma: no cover
            f"WwwAuthenticate header does not seems to contain a valid content. No challenge detected."
        )


class XDnsPrefetchControl(CustomHeader):
    """
    The X-DNS-Prefetch-Control HTTP response header controls DNS prefetching, a feature by which browsers proactively
    perform domain name resolution on both links that the user may choose to follow as well as URLs
    for items referenced by the document, including images, CSS, JavaScript, and so forth.
    """

    __tags__ = ["response"]

    def __init__(self, enable: bool = True, **kwargs: Optional[str]):
        """
        :param enable: Toggle the specified behaviour.
        :param kwargs:
        """
        super().__init__("on" if enable else "off", **kwargs)


class Location(CustomHeader):
    """
    The Location response header indicates the URL to redirect a page to.
    It only provides a meaning when served with a 3xx (redirection) or 201 (created) status response.
    """

    __tags__ = ["response"]

    def __init__(self, uri: str, **kwargs: Optional[str]):
        """
        :param uri: A relative (to the request URL) or absolute URL.
        :param kwargs:
        """
        super().__init__(uri, **kwargs)


class From(CustomHeader):
    """
    The From request header contains an Internet email address for a human user who controls the requesting user agent.
    If you are running a robotic user agent (e.g. a crawler), the From header should be sent, so you can be contacted
    if problems occur on servers, such as if the robot is sending excessive, unwanted, or invalid requests.
    """

    __tags__: List[str] = ["request"]

    def __init__(self, email: str, **kwargs: Optional[str]):
        """
        :param email: A machine-usable email address. See RFC 5322.
        :param kwargs:
        """
        if (
            fullmatch(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)", email)
            is None
        ):
            raise ValueError(  # pragma: no cover
                f"'{email}' is not a valid RFC 5322 email."
            )

        super().__init__(email, **kwargs)


class ContentRange(CustomHeader):
    """
    The Content-Range response HTTP header indicates where in a full body message a partial message belongs.
    """

    __tags__ = ["response"]

    def __init__(
        self,
        unit: str,
        start: int,
        end: int,
        size: Union[str, int],
        **kwargs: Optional[str],
    ):
        """
        :param unit: The unit in which ranges is specified. This is usually bytes.
        :param start: An integer in the given unit indicating the beginning of the request range.
        :param end: An integer in the given unit indicating the end of the requested range.
        :param size: The total size of the document (or '*' if unknown).
        :param kwargs:
        >>> header = ContentRange("bytes", 0, 1024, 4096)
        >>> repr(header)
        'Content-Range: bytes 0-1024/4096'
        >>> header.get_size()
        4096
        >>> header.unpack()
        ('bytes', '0', '1024', '4096')
        """
        super().__init__(f"{unit} {start}-{end}/{size}", **kwargs)

    def unpack(self) -> Tuple[str, str, str, str]:
        """Provide a basic way to parse ContentRange format."""
        return findall(
            r"^([0-9a-zA-Z*]+) ([0-9a-zA-Z*]+)-([0-9a-zA-Z*]+)/([0-9a-zA-Z*]+)$",
            str(self),
        )[0]

    def get_unit(self) -> str:
        """Retrieve the unit in which ranges is specified."""
        return self.unpack()[0]

    def get_start(self) -> int:
        """Get the beginning of the request range."""
        return int(self.unpack()[1])

    def get_end(self) -> int:
        """Get the end of the requested range."""
        return int(self.unpack()[2])

    def get_size(self) -> Union[str, int]:
        """Get the total size of the document (or '*' if unknown)."""
        size: str = self.unpack()[3]
        return int(size) if size.isdigit() else size


class CacheControl(CustomHeader):
    """
    The Cache-Control HTTP header holds directives (instructions) for caching in
    both requests and responses. A given directive in a request does not mean the
    same directive should be in the response.
    """

    __tags__ = ["request", "response"]
    __squash__ = True

    def __init__(
        self,
        directive: Optional[str] = None,
        max_age: Optional[int] = None,
        max_stale: Optional[int] = None,
        min_fresh: Optional[int] = None,
        s_maxage: Optional[int] = None,
        **kwargs: Optional[str],
    ):
        """
        Pass only one parameter per CacheControl instance.
        :param directive: Could be one of must-revalidate, no-cache, no-store, no-transform, public, private, proxy-revalidate, only-if-cached, no-transform.
        :param max_age: The maximum amount of time a resource is considered fresh. Unlike Expires, this directive is relative to the time of the request.
        :param max_stale: Indicates the client will accept a stale response. An optional value in seconds indicates the upper limit of staleness the client will accept.
        :param min_fresh: Indicates the client wants a response that will still be fresh for at least the specified number of seconds.
        :param s_maxage: Overrides max-age or the Expires header, but only for shared caches (e.g., proxies). Ignored by private caches.
        :param kwargs:
        """
        if [
            directive is not None,
            max_age is not None,
            max_stale is not None,
            min_fresh is not None,
            s_maxage is not None,
        ].count(True) != 1:
            raise ValueError(  # pragma: no cover
                "You should only pass one parameter to a single CacheControl instance."
            )

        args: Dict = {
            "max-age": max_age,
            "max-stale": max_stale,
            "min-fresh": min_fresh,
            "s-maxage": s_maxage,
        }

        args.update(kwargs)

        super().__init__(directive if directive is not None else "", **args)


class Expires(Date):
    """
    The Expires header contains the date/time after which the response is considered stale.
    Invalid dates, like the value 0, represent a date in the past and mean that the resource is already expired.
    """

    __tags__ = ["response"]

    def __init__(
        self, datetime_or_custom: Union[datetime, str], **kwargs: Optional[str]
    ):
        super().__init__(datetime_or_custom, **kwargs)


class IfModifiedSince(Date):
    """
    The If-Modified-Since request HTTP header makes the request conditional
    """

    __tags__ = ["request"]

    def __init__(self, dt: Union[datetime, str], **kwargs: Optional[str]):
        """
        :param dt:
        :param kwargs:
        """
        super().__init__(dt, **kwargs)


class IfUnmodifiedSince(Date):
    """
    The If-Unmodified-Since request HTTP header makes the request conditional
    """

    def __init__(self, dt: Union[datetime, str], **kwargs: Optional[str]):
        """
        :param dt:
        :param kwargs:
        """
        super().__init__(dt, **kwargs)


class KeepAlive(CustomHeader):
    """
    The Keep-Alive general header allows the sender to hint about how the connection may be used to
    set a timeout and a maximum amount of requests.
    """

    __squash__ = True
    __tags__ = ["request", "response"]

    def __init__(
        self,
        timeout: Optional[int] = None,
        max_: Optional[int] = None,
        **kwargs: Optional[str],
    ):
        """
        :param timeout: indicating the minimum amount of time an idle connection has to be kept opened (in seconds).
        :param max: indicating the maximum number of requests that can be sent on this connection before closing it.
        :param kwargs:
        """
        if timeout is not None and max_ is not None:
            raise ValueError(  # pragma: no cover
                "Can only provide one parameter per KeepAlive instance, either timeout or max, not both."
            )

        args: Dict = {"timeout": timeout, "max": max_}

        args.update(kwargs)

        super().__init__("", **args)


class IfMatch(CustomHeader):
    """
    The If-Match HTTP request header makes the request conditional. For GET and HEAD methods,
    the server will send back the requested resource only if it matches one of the listed ETags.
    For PUT and other non-safe methods, it will only upload the resource in this case.
    """

    __squash__ = True
    __tags__ = ["request"]

    def __init__(self, etag_value: str, **kwargs: Optional[str]):
        """
        :param etag_value: Entity tags uniquely representing the requested resources. They are a string of ASCII characters placed between double quotes (like "675af34563dc-tr34").
        :param kwargs:
        """
        super().__init__(quote(etag_value), **kwargs)


class IfNoneMatch(IfMatch):
    """
    The If-None-Match HTTP request header makes the request conditional. For GET and HEAD methods,
    the server will send back the requested resource, with a 200 status, only if it doesn't have an ETag matching
    the given ones. For other methods, the request will be processed only if the eventually existing resource's
    ETag doesn't match any of the values listed.
    """

    def __init__(self, etag_value: str, **kwargs: Optional[str]):
        super().__init__(etag_value, **kwargs)


class Server(CustomHeader):
    """The Server header describes the software used by the origin server that handled the request —
    that is, the server that generated the response."""

    __tags__ = ["response"]

    def __init__(self, product: str, **kwargs: Optional[str]):
        """
        :param product: The name of the software or product that handled the request. Usually in a format similar to User-Agent.
        :param kwargs:
        """
        super().__init__(product, **kwargs)


class Vary(CustomHeader):
    """The Vary HTTP response header determines how to match future request headers to decide whether a cached response
    can be used rather than requesting a fresh one from the origin server."""

    __squash__ = True
    __tags__ = ["response"]

    def __init__(self, header_name: str, **kwargs: Optional[str]):
        """
        :param header_name: An header name to take into account when deciding whether or not a cached response can be used.
        :param kwargs:
        """
        super().__init__(prettify_header_name(header_name), **kwargs)
